"""
Data attributes for events within the architecture subdomain `learning`.

These attributes follow the form of attr objects specified in OEP-49 data
pattern.
"""
import json
import socket
from datetime import datetime, timezone
from uuid import UUID, uuid1

import attr
import attrs
from django.conf import settings

import openedx_events


def _ensure_utc_time(_, attribute, value):
    """
    Ensure the value is a UTC datetime.

    Note: Meant to be used along-side an instance_of attr validator.
    """
    if value.tzinfo and value.tzinfo == timezone.utc:
        return
    raise ValueError(f"'{attribute.name}' must have timezone.utc")


def get_service_name():
    """
    Get the service name of the producing/consuming service of an event (or None if not set).

    Uses EVENTS_SERVICE_NAME setting if present, otherwise looks for SERVICE_VARIANT.
    """
    # .. setting_name: EVENTS_SERVICE_NAME
    # .. setting_default: None
    # .. setting_description: Identifier for the producing/consuming service of an event. For example, "cms" or
    #   "course-discovery." Used, among other places, to determine the source header of the event.
    return getattr(settings, "EVENTS_SERVICE_NAME", None) or getattr(settings, "SERVICE_VARIANT", None)


def _get_source():
    """
    Get the source for an event using the service name.

    If the service name is set, the full source will be set to openedx/<service_name>/web or
    openedx/SERVICE_NAME_UNSET/web if service name is None.
    """
    return "openedx/{service}/web".format(service=(get_service_name() or "SERVICE_NAME_UNSET"))


@attr.s(frozen=True)
class EventsMetadata:
    """
    Attributes defined for Open edX Events metadata object.

    The attributes defined in this class are a subset of the
    OEP-41: Asynchronous Server Event Message Format.

    Arguments:
        id (UUID): event identifier.
        event_type (str): name of the event.
        minorversion (int): (optional) version of the event type. Defaults to 0.
        source (str): logical source of an event.
        sourcehost (str): physical source of the event.
        time (datetime): (optional) timestamp when the event was sent with
            UTC timezone. Defaults to current time in UTC. See OEP-41 for
            details.
        sourcelib (tuple of ints): Open edX Events library version. A tuple was
            selected so that version comparisons don't have to worry about
            lexical ordering of strings (e.g. '0.9.0' vs. '0.10.0').
    """

    event_type = attr.ib(type=str, validator=attr.validators.instance_of(str))
    id = attr.ib(
        type=UUID, default=None,
        converter=attr.converters.default_if_none(attr.Factory(lambda: uuid1())),  # pylint: disable=unnecessary-lambda
        validator=attr.validators.instance_of(UUID),
    )
    minorversion = attr.ib(
        type=int, default=None,
        converter=attr.converters.default_if_none(0), validator=attr.validators.instance_of(int)
    )
    source = attr.ib(
        type=str, default=None,
        converter=attr.converters.default_if_none(attr.Factory(_get_source)),
        validator=attr.validators.instance_of(str),
    )
    sourcehost = attr.ib(
        type=str, default=None,
        converter=attr.converters.default_if_none(
            attr.Factory(lambda: socket.gethostname())  # pylint: disable=unnecessary-lambda
        ),
        validator=attr.validators.instance_of(str),
    )
    time = attr.ib(
        type=datetime, default=None,
        converter=attr.converters.default_if_none(attr.Factory(lambda: datetime.now(timezone.utc))),
        validator=[attr.validators.instance_of(datetime), _ensure_utc_time],
    )
    sourcelib = attr.ib(
        type=tuple, default=None,
        converter=attr.converters.default_if_none(
            attr.Factory(lambda: tuple(map(int, openedx_events.__version__.split("."))))
        ),
        validator=attr.validators.instance_of(tuple),
    )

    def to_json_data(self):
        """
        Create a json-compatible dictionary of the instance.
        """
        def value_serializer(inst, field, value):  # pylint: disable="unused-argument"
            if isinstance(value, UUID):
                return str(value)
            elif isinstance(value, datetime):
                return value.isoformat()
            else:
                return value
        return attrs.asdict(self, value_serializer=value_serializer)

    def to_json(self):
        """
        Serialize instance to json string.
        """
        return json.dumps(self.to_json_data())

    @classmethod
    def from_json(cls, json_string):
        """
        Create an instance from a json string.

        Arguments:
            json_string (str): A json representation of an EventsMetadata object, created with as_json_string
        Returns:
            An EventsMetadata object
        """
        as_json = json.loads(json_string)
        time = datetime.fromisoformat(as_json['time'])
        sourcelib = tuple(as_json['sourcelib'])
        return cls(event_type=as_json['event_type'], id=UUID(as_json['id']), source=as_json['source'],
                   sourcehost=as_json['sourcehost'], time=time, sourcelib=sourcelib)



"""
Custom exceptions thrown by Open edX events tooling.
"""


class OpenEdxEventException(Exception):
    """
    Base class for Open edX Events exceptions.
    """

    def __init__(self, message=""):
        """
        Init method for OpenEdxEventException base class.

        Arguments:
            message (str): message describing why the exception was raised.
        """
        super().__init__()
        self.message = message

    def __str__(self):
        """
        Show string representation of OpenEdxEventException using its message.
        """
        return self.message


class InstantiationError(OpenEdxEventException):
    """
    Describes errors that occur while instantiating events.

    This exception is raised when there's an error instantiating an Open edX
    event, it can be that a required argument for the event definition is
    missing.
    """

    def __init__(self, event_type="", message=""):
        """
        Init method for InstantiationError custom exception class.

        Arguments:
            event_type (str): name of the event raising the exception.
            message (str): message describing why the exception was raised.
        """
        super().__init__(
            message="InstantiationError {event_type}: {message}".format(
                event_type=event_type, message=message
            )
        )


class SenderValidationError(OpenEdxEventException):
    """
    Describes errors that occur while validating arguments of send methods.
    """

    def __init__(self, event_type="", message=""):
        """
        Init method for SenderValidationError custom exception class.

        Arguments:
            event_type (str): name of the event raising the exception.
            message (str): message describing why the exception was raised.
        """
        super().__init__(
            message="SenderValidationError {event_type}: {message}".format(
                event_type=event_type, message=message
            )
        )


class ProducerConfigurationError(OpenEdxEventException):
    """
    Describes errors that occurs while validating format of producer signal configuration.
    """

    def __init__(self, event_type="", message=""):
        """
        Init method for ProducerConfigurationError custom exception class.

        Arguments:
            event_type (str): name of the event raising the exception.
            message (str): message describing why the exception was raised.
        """
        super().__init__(
            message="ProducerConfigurationError {event_type}: {message}".format(
                event_type=event_type, message=message
            )
        )



"""
Management command to generate and store Avro schemas for testing.
"""
import json
import logging
import os
from importlib import import_module

from django.core.management.base import BaseCommand

from openedx_events.event_bus.avro.serializer import AvroSignalSerializer
from openedx_events.tooling import KNOWN_UNSERIALIZABLE_SIGNALS, OpenEdxPublicSignal, load_all_signals

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Management command to save current Avro schemas for all signals for schema evolution testing.
    """

    help = """
    Generate and save the current Avro schemas of OpenEdxPublicSignals.

    Schemas will be saved in openedx_events/avro/tests/schemas. The folder will be created if it does not already exist.
    Should only be used for new signals. Even if a schema is changing, it is recommended you leave the original
    in place. Only overwrite the file for an exceptional case, like the schema still being under development.

    Example::

        # one signal
        python manage.py generate_avro_schemas org.openedx.learning.course.enrollment.changed.v1

        # multiple signals
        python manage.py generate_avro_schemas org.openedx.learning.course.enrollment.changed.v1 /
            org.openedx.learning.course.unenrollment.completed.v1

        # all signals
        python manage.py generate_avro_schemas --all

    """

    def add_arguments(self, parser):
        """
        Add arguments for either individual event types or all of them.
        """
        parser.add_argument(
            'types',
            nargs='*',
            type=str,
            help='Event types for which to write and save the schema, separated by a space'
        )

        parser.add_argument(
            '--all',
            action="store_true",
            help='Write schema for all event types'
        )

    def handle(self, *args, **options):
        """
        Create consumer based on django settings and consume events.
        """
        load_all_signals()
        if options['all']:
            signals = OpenEdxPublicSignal.all_events()
        else:
            signals = [OpenEdxPublicSignal.get_signal_by_type(event_type) for event_type in options['types']]

        for signal in signals:
            if signal.event_type in KNOWN_UNSERIALIZABLE_SIGNALS:
                logger.info(f"Known unserializable signal: {signal.event_type}. Skipping.")
                continue
            serializer = AvroSignalSerializer(signal)
            schema_dict = serializer.schema
            filename = f"{signal.event_type.replace('.','+')}_schema.avsc"
            root_path = import_module('openedx_events').__path__[0]
            folder_path = f"{root_path}/event_bus/avro/tests/schemas"
            full_file_name = f"{folder_path}/{filename}"
            if os.path.exists(full_file_name):
                confirmation = input(f"Warning: overwriting schema for {signal.event_type}. It is recommended to leave"
                                     f" existing schemas unchanged. Are you sure you want to "
                                     f"continue [y/n]? ")
                if confirmation.lower().strip() != 'y':
                    logger.info(f"Skipping generating schema for {signal.event_type}")
                    continue
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
            logger.info(f"Writing {full_file_name}")
            with open(full_file_name, 'w') as writes:
                writes.write(json.dumps(schema_dict, indent=2))



"""
Makes ``consume_events`` management command available.
"""
import json
import logging

from django.core.management.base import BaseCommand

from openedx_events.event_bus import make_single_consumer
from openedx_events.tooling import load_all_signals

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Management command for consumer workers in the event bus.
    """

    help = """
    Consume messages from a topic and emit their data with the correct signal.

    Example::

        python manage.py consume_events -t user-login -g user-activity-service

        # send extra args, for example pass check_backlog flag to redis consumer
        python manage.py cms consume_events -t user-login -g user-activity-service  --extra '{"check_backlog": true}'

        # send extra args, for example replay events from specific redis msg id.
        python manage.py cms consume_events -t user-login -g user-activity-service \
            --extra '{"last_read_msg_id": "1679676448892-0"}'
    """

    def add_arguments(self, parser):
        """
        Add arguments for parsing topic, group, and extra args.
        """
        parser.add_argument(
            '-t', '--topic',
            nargs=1,
            required=True,
            help='Topic to consume (without environment prefix)'
        )
        parser.add_argument(
            '-g', '--group_id',
            nargs=1,
            required=True,
            help='Consumer group id'
        )
        parser.add_argument(
            '--extra',
            nargs='?',
            type=str,
            required=False,
            help='JSON object to pass additional arguments to the consumer.'
        )

    def handle(self, *args, **options):
        """
        Create consumer based on django settings and consume events.
        """
        try:
            # load additional arguments specific for the underlying implementation of event_bus.
            extra = json.loads(options.get('extra') or '{}')
            load_all_signals()
            event_consumer = make_single_consumer(
                topic=options['topic'][0],
                group_id=options['group_id'][0],
                **extra,
            )
            event_consumer.consume_indefinitely()
        except Exception:  # pylint: disable=broad-except
            logger.exception("Error consuming events")



"""
Standardized signals definitions for events within the architecture subdomain ``learning``.

All signals defined in this module must follow the name and versioning
conventions specified in docs/decisions/0002-events-naming-and-versioning.rst

They also must comply with the payload definition specified in
docs/decisions/0003-events-payload.rst
"""

from openedx_events.learning.data import (
    CertificateData,
    CohortData,
    CourseAccessRoleData,
    CourseDiscussionConfigurationData,
    CourseEnrollmentData,
    CourseNotificationData,
    DiscussionThreadData,
    ExamAttemptData,
    ManageStudentsPermissionData,
    PersistentCourseGradeData,
    ProgramCertificateData,
    UserData,
    UserNotificationData,
    XBlockSkillVerificationData,
)
from openedx_events.tooling import OpenEdxPublicSignal

# .. event_type: org.openedx.learning.student.registration.completed.v1
# .. event_name: STUDENT_REGISTRATION_COMPLETED
# .. event_description: emitted when the user registration process in the LMS is completed.
# .. event_data: UserData
STUDENT_REGISTRATION_COMPLETED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.student.registration.completed.v1",
    data={
        "user": UserData,
    }
)


# .. event_type: org.openedx.learning.auth.session.login.completed.v1
# .. event_name: SESSION_LOGIN_COMPLETED
# .. event_description: emitted when the user's login process in the LMS is completed.
# .. event_data: UserData
SESSION_LOGIN_COMPLETED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.auth.session.login.completed.v1",
    data={
        "user": UserData,
    }
)


# .. event_type: org.openedx.learning.course.enrollment.created.v1
# .. event_name: COURSE_ENROLLMENT_CREATED
# .. event_description: emitted when the user's enrollment process is completed.
# .. event_data: CourseEnrollmentData
COURSE_ENROLLMENT_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.enrollment.created.v1",
    data={
        "enrollment": CourseEnrollmentData,
    }
)


# .. event_type: org.openedx.learning.course.enrollment.changed.v1
# .. event_name: COURSE_ENROLLMENT_CHANGED
# .. event_description: emitted when the user's enrollment update process is completed.
# .. event_data: CourseEnrollmentData
COURSE_ENROLLMENT_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.enrollment.changed.v1",
    data={
        "enrollment": CourseEnrollmentData,
    }
)


# .. event_type: org.openedx.learning.course.unenrollment.completed.v1
# .. event_name: COURSE_UNENROLLMENT_COMPLETED
# .. event_description: emitted when the user's unenrollment process is completed.
# .. event_data: CourseEnrollmentData
COURSE_UNENROLLMENT_COMPLETED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.unenrollment.completed.v1",
    data={
        "enrollment": CourseEnrollmentData,
    }
)


# .. event_type: org.openedx.learning.certificate.created.v1
# .. event_name: CERTIFICATE_CREATED
# .. event_description: emitted when the user's certificate creation process is completed.
# .. event_data: CertificateData
CERTIFICATE_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.certificate.created.v1",
    data={
        "certificate": CertificateData,
    }
)

# .. event_type: org.openedx.learning.program.certificate.awarded.v1
# .. event_name: PROGRAM_CERTIFICATE_AWARDED
# .. event_description: Emit when a program certificate is awarded to a learner
# .. event_data: ProgramCertificateData
PROGRAM_CERTIFICATE_AWARDED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.program.certificate.awarded.v1",
    data={
        "program_certificate": ProgramCertificateData,
    }
)

# .. event_type: org.openedx.learning.certificate.changed.v1
# .. event_name: CERTIFICATE_CHANGED
# .. event_description: emitted when the user's certificate update process is completed.
# .. event_data: CertificateData
CERTIFICATE_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.certificate.changed.v1",
    data={
        "certificate": CertificateData,
    }
)


# .. event_type: org.openedx.learning.certificate.revoked.v1
# .. event_name: CERTIFICATE_REVOKED
# .. event_description: emitted when the user's certificate annulation process is completed.
# .. event_data: CertificateData
CERTIFICATE_REVOKED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.certificate.revoked.v1",
    data={
        "certificate": CertificateData,
    }
)

# .. event_type: org.openedx.learning.program.certificate.revoked.v1
# .. event_name: PROGRAM_CERTIFICATE_REVOKED
# .. event_description: Emit when a program certificate is revoked from a learner
# .. event_data: ProgramCertificateData
PROGRAM_CERTIFICATE_REVOKED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.program.certificate.revoked.v1",
    data={
        "program_certificate": ProgramCertificateData,
    }
)

# .. event_type: org.openedx.learning.cohort_membership.changed.v1
# .. event_name: COHORT_MEMBERSHIP_CHANGED
# .. event_description: emitted when the user's cohort update is completed.
# .. event_data: CohortData
COHORT_MEMBERSHIP_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.cohort_membership.changed.v1",
    data={
        "cohort": CohortData,
    }
)


# .. event_type: org.openedx.learning.discussions.configuration.changed.v1
# .. event_name: COURSE_DISCUSSIONS_CHANGED
# .. event_description: emitted when the configuration for a course's discussions changes in the course
#       Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
# .. event_data: CourseDiscussionConfigurationData
COURSE_DISCUSSIONS_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.discussions.configuration.changed.v1",
    data={
        "configuration": CourseDiscussionConfigurationData
    }
)

# .. event_type: org.openedx.learning.course.persistent_grade.summary.v1
# .. event_name: PERSISTENT_GRADE_SUMMARY_CHANGED
# .. event_description: emitted when a grade changes in the course
# .. event_data: PersistentCourseGradeData
PERSISTENT_GRADE_SUMMARY_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.persistent_grade_summary.changed.v1",
    data={
        "grade": PersistentCourseGradeData,
    }
)


# .. event_type: org.openedx.learning.xblock.skill.verified.v1
# .. event_name: XBLOCK_SKILL_VERIFIED
# .. event_description: Fired when an XBlock skill is verified.
# .. event_data: XBlockSkillVerificationData
XBLOCK_SKILL_VERIFIED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.xblock.skill.verified.v1",
    data={
        "xblock_info": XBlockSkillVerificationData,
    }
)

# .. event_type: org.openedx.learning.user.notification.requested.v1
# .. event_name: USER_NOTIFICATION_REQUESTED
# .. event_description: Can be fired from apps to send user notifications.
# .. event_data: UserNotificationSendListData
# Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
#
USER_NOTIFICATION_REQUESTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.notification.requested.v1",
    data={
        "notification_data": UserNotificationData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.submitted.v1
# .. event_name: EXAM_ATTEMPT_SUBMITTED
# .. event_description: Emitted when an exam attempt is submitted by a learner in edx-exams.
# .. event_data: ExamAttemptData
EXAM_ATTEMPT_SUBMITTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.submitted.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.rejected.v1
# .. event_name: EXAM_ATTEMPT_REJECTED
# .. event_description: Emitted when an exam attempt is marked rejected in edx-exams.
# .. event_data: ExamAttemptData
EXAM_ATTEMPT_REJECTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.rejected.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.verified.v1
# .. event_name: EXAM_ATTEMPT_VERIFIED
# .. event_description: Emitted when an exam attempt is marked verified in edx-exams.
# .. event_data: ExamAttemptData
EXAM_ATTEMPT_VERIFIED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.verified.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.errored.v1
# .. event_name: EXAM_ATTEMPT_ERRORED
# .. event_description: Emitted when a learner's exam attempt errors out in edx-exams.
# .. event_data: ExamAttemptData
EXAM_ATTEMPT_ERRORED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.errored.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.exam.attempt.reset.v1
# .. event_name: EXAM_ATTEMPT_RESET
# .. event_description: Emitted when an exam attempt is reset in edx-exams.
# .. event_data: ExamAttemptData
EXAM_ATTEMPT_RESET = OpenEdxPublicSignal(
    event_type="org.openedx.learning.exam.attempt.reset.v1",
    data={
        "exam_attempt": ExamAttemptData,
    }
)

# .. event_type: org.openedx.learning.user.course_access_role.added.v1
# .. event_name: COURSE_ACCESS_ROLE_ADDED
# .. event_description: Emitted when a user is given a course access role.
# .. event_data: CourseAccessRoleData
COURSE_ACCESS_ROLE_ADDED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.course_access_role.added.v1",
    data={
        "course_access_role_data": CourseAccessRoleData,
    }
)

# .. event_type: org.openedx.learning.user.course_access_role.removed.v1
# .. event_name: COURSE_ACCESS_ROLE_REMOVED
# .. event_description: Emitted when a course access role is removed from a user.
# .. event_data: CourseAccessRoleData
COURSE_ACCESS_ROLE_REMOVED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.course_access_role.removed.v1",
    data={
        "course_access_role_data": CourseAccessRoleData,
    }
)

# .. event_type: org.openedx.learning.user.manage.students.permission.added.v1
# .. event_name: MANAGE_STUDENTS_PERMISSION_ADDED
# .. event_description: Emitted when permission to manage students within a course is given to a user.
# .. event_data: ManageStudentsPermissionData
MANAGE_STUDENTS_PERMISSION_ADDED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.manage.students.permission.added.v1",
    data={
        "course_staff_data": ManageStudentsPermissionData,
    }
)

# .. event_type: org.openedx.learning.user.manage.students.permission.removed.v1
# .. event_name: MANAGE_STUDENTS_PERMISSION_REMOVED
# .. event_description: Emitted when permission to manage students within a course is removed from a user.
# .. event_data: ManageStudentsPermissionData
MANAGE_STUDENTS_PERMISSION_REMOVED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.user.manage.students.permission.removed.v1",
    data={
        "course_staff_data": ManageStudentsPermissionData,
    }
)

# .. event_type: org.openedx.learning.forum.thread.created.v1
# .. event_name: FORUM_THREAD_CREATED
# .. event_description: Emitted when a new thread is created in a discussion
# .. event_data: DiscussionThreadData
#       Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
FORUM_THREAD_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.thread.created.v1",
    data={
        "thread": DiscussionThreadData,
    }
)

# .. event_type: org.openedx.learning.forum.thread.response.created.v1
# .. event_name: FORUM_THREAD_RESPONSE_CREATED
# .. event_description: Emitted when a new response is added to a thread
# .. event_data: DiscussionThreadData
#        Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
FORUM_THREAD_RESPONSE_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.response.created.v1",
    data={
        "thread": DiscussionThreadData,
    }
)

# .. event_type: org.openedx.learning.forum.thread.response.comment.created.v1
# .. event_name: FORUM_RESPONSE_COMMENT_CREATED
# .. event_description: Emitted when a new comment is added to a response
# .. event_data: DiscussionThreadData
#       Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
FORUM_RESPONSE_COMMENT_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.response.created.v1",
    data={
        "thread": DiscussionThreadData,
    }
)


# .. event_type: org.openedx.learning.course.notification.requested.v1
# .. event_name: COURSE_NOTIFICATION_REQUESTED
# .. event_description: Emitted when a notification is requested for a course
# .. event_data: CourseNotificationData
# Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
#
COURSE_NOTIFICATION_REQUESTED = OpenEdxPublicSignal(
    event_type="org.openedx.learning.course.notification.requested.v1",
    data={
        "course_notification_data": CourseNotificationData,
    }
)



"""
Standardized signals definitions for events within the architecture subdomain ``analytics``.

All signals defined in this module must follow the name and versioning
conventions specified in OEP-41.

They also must comply with the payload definition specified in
docs/decisions/0003-events-payload.rst
"""

from openedx_events.analytics.data import TrackingLogData
from openedx_events.tooling import OpenEdxPublicSignal

# .. event_type: org.openedx.analytics.tracking.event.emitted.v1
# .. event_name: TRACKING_EVENT_EMITTED
# .. event_description: emitted when a tracking log is created.
# .. event_data: TrackingLogData
TRACKING_EVENT_EMITTED = OpenEdxPublicSignal(
    event_type="org.openedx.analytics.tracking.event.emitted.v1",
    data={
        "tracking_log": TrackingLogData,
    }
)



"""Test interplay of the various Avro helper classes"""
import io
import os
from datetime import datetime
from typing import List
from unittest import TestCase

from fastavro import schemaless_reader, schemaless_writer
from fastavro.repository.base import SchemaRepositoryError
from fastavro.schema import load_schema
from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2, LibraryUsageLocatorV2

from openedx_events.event_bus.avro.deserializer import AvroSignalDeserializer, deserialize_bytes_to_event_data
from openedx_events.event_bus.avro.serializer import AvroSignalSerializer, serialize_event_data_to_bytes
from openedx_events.event_bus.avro.tests.test_utilities import (
    EventData,
    NestedAttrsWithDefaults,
    SimpleAttrsWithDefaults,
    SubTestData0,
    SubTestData1,
    create_simple_signal,
)
from openedx_events.tests.utils import FreezeSignalCacheMixin
from openedx_events.tooling import KNOWN_UNSERIALIZABLE_SIGNALS, OpenEdxPublicSignal, load_all_signals


def generate_test_data_for_schema(schema):  # pragma: no cover
    """
    Generates a test data dict for the given schema.

    Arguments:
        schema: A JSON representation of an Avro schema

    Returns:
         A dictionary of test data parseable by the schema
    """
    defaults_per_type = {
        'long': 1,
        'boolean': True,
        'string': "default",
        'double': 1.0,
        'null': None,
    }

    def get_default_value_or_raise(schema_field_type):
        try:
            return defaults_per_type[schema_field_type]
        # 'None' is the default value for type=null so we can't just check if default_value is not None
        except KeyError as exc:
            raise Exception(f"Unsupported type {schema_field_type}") from exc  # pylint: disable=broad-exception-raised

    data_dict = {}
    top_level = schema['fields']
    for field in top_level:
        key = field['name']
        field_type = field['type']

        # some fields (like optional ones) accept multiple types. Choose the first one and run with it.
        if isinstance(field_type, list):
            field_type = field_type[0]

        # if the field_type is a dict, we're either dealing with a list or a custom object
        if isinstance(field_type, dict):
            sub_field_type = field_type['type']
            if sub_field_type == "array":
                # if we're dealing with a list, "items" will be the type of items in the list
                data_dict.update({key: [get_default_value_or_raise(field_type['items'])]})
            elif sub_field_type == "record":
                # if we're dealing with a record, recurse into the record
                data_dict.update({key: generate_test_data_for_schema(field_type)})
            else:
                raise Exception(f"Unsupported type {field_type}")  # pylint: disable=broad-exception-raised

        # a record is a collection of fields rather than a field itself, so recursively generate and add each field
        elif field_type == "record":
            data_dict.update([generate_test_data_for_schema(sub_field) for sub_field in field['fields']])
        else:
            data_dict.update({key: get_default_value_or_raise(field_type)})

    return data_dict


def generate_test_event_data_for_data_type(data_type):  # pragma: no cover
    """
    Generates test data for use in the event bus test cases.

    Builds data by filling in dummy data for basic data types (int/float/bool/str)
    and recursively breaks down the classes for nested classes into basic data types.

    Arguments:
        data_type: The type of the data which we are generating data for

    Returns:
        (dict): A data dictionary containing dummy data for all attributes of the class
    """
    defaults_per_type = {
        int: 1,
        bool: True,
        str: "default",
        float: 1.0,
        CourseKey: CourseKey.from_string("course-v1:edX+DemoX.1+2014"),
        UsageKey: UsageKey.from_string(
            "block-v1:edx+DemoX+Demo_course+type@video+block@UaEBjyMjcLW65gaTXggB93WmvoxGAJa0JeHRrDThk",
        ),
        LibraryLocatorV2: LibraryLocatorV2.from_string('lib:MITx:reallyhardproblems'),
        LibraryUsageLocatorV2: LibraryUsageLocatorV2.from_string('lb:MITx:reallyhardproblems:problem:problem1'),
        List[int]: [1, 2, 3],
        datetime: datetime.now(),
    }
    data_dict = {}
    for attribute in data_type.__attrs_attrs__:
        result = defaults_per_type.get(attribute.type, None)
        if result is not None:
            data_dict.update({attribute.name: result})
        elif attribute.type in [dict, list]:
            # pylint: disable-next=broad-exception-raised
            raise Exception("Unable to generate Avro schema for dict or array fields")
        else:
            data_dict.update({attribute.name: attribute.type(**generate_test_event_data_for_data_type(attribute.type))})
    return data_dict


def generate_test_data_for_signal(signal: OpenEdxPublicSignal) -> dict:
    """
    Generates test data for use in the event bus test cases.

    Builds data by filling in dummy data for basic data types (int/float/bool/str)
    and recursively breaks down the classes for nested classes into basic data types.

    Arguments:
        data_type: The type of the data which we are generating data for

    Returns:
        (dict): A data dictionary containing dummy data for all attributes of the class
    """
    test_data = {}
    for key, curr_class in signal.init_data.items():
        example_data = generate_test_event_data_for_data_type(curr_class)
        example_data_processed = curr_class(**example_data)
        test_data.update({key: example_data_processed})
    return test_data


class TestAvro(FreezeSignalCacheMixin, TestCase):
    """Tests for end-to-end serialization and deserialization of events and schema evolution"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Ensure we can usefully call all_events()
        load_all_signals()

    def test_all_events(self):
        for signal in OpenEdxPublicSignal.all_events():
            if signal.event_type in KNOWN_UNSERIALIZABLE_SIGNALS:
                continue
            test_data = generate_test_data_for_signal(signal)
            serializer = AvroSignalSerializer(signal)
            serialized = serializer.to_dict(test_data)
            deserializer = AvroSignalDeserializer(signal)
            deserialized = deserializer.from_dict(serialized)
            self.assertDictEqual(deserialized, test_data)

    def test_evolution_is_forward_compatible(self):
        """
        Test current version of events is forward compatible with stored schemas.

        There's no assert because as long as the test doesn't raise an error, the new schema is
        forward-compatible with the original.
        """
        for signal in OpenEdxPublicSignal.all_events():
            if signal.event_type in KNOWN_UNSERIALIZABLE_SIGNALS:
                continue
            test_data = generate_test_data_for_signal(signal)
            serializer = AvroSignalSerializer(signal)
            schema_dict = serializer.schema

            # write to bytes using current schema
            current_out = io.BytesIO()
            data_dict = serializer.to_dict(test_data)
            schemaless_writer(current_out, schema_dict, data_dict)
            current_out.seek(0)
            current_event_bytes = current_out.read()

            # get stored schema
            schema_filename = f"{os.path.dirname(os.path.abspath(__file__))}/schemas/" \
                              f"{signal.event_type.replace('.','+')}_schema.avsc"
            try:
                stored_schema = load_schema(schema_filename)
            except SchemaRepositoryError:  # pragma: no cover
                self.fail(f"Missing file {schema_filename}. If a new signal has been added, you may need to run the"
                          f" generate_avro_schemas management command to save the signal schema.")

            data_file_current = io.BytesIO(current_event_bytes)

            # read bytes using stored schema
            schemaless_reader(data_file_current, reader_schema=stored_schema, writer_schema=schema_dict)

    def test_evolution_is_backward_compatible(self):
        """
        Test current version of events is backward compatible with stored schemas.

        There's no assert because as long as the test doesn't raise an error, the new schema is
        backward-compatible with the original.

        Caveat: While we can go from an attrs class to a schema, we cannot go from a schema to an attrs class in the
        same way because of custom serializers. Thus, when we are generating test data, we can only generate
        the dictionary we would get after calling schema.to_dict.
        """
        for signal in OpenEdxPublicSignal.all_events():
            if signal.event_type in KNOWN_UNSERIALIZABLE_SIGNALS:
                continue
            serializer = AvroSignalSerializer(signal)
            schema_dict = serializer.schema

            # get stored schema
            schema_filename = f"{os.path.dirname(os.path.abspath(__file__))}/schemas/" \
                              f"{signal.event_type.replace('.','+')}_schema.avsc"
            try:
                old_schema = load_schema(schema_filename)
            except SchemaRepositoryError:  # pragma: no cover
                self.fail(f"Missing file {schema_filename}. If a new signal has been added, you may need to run the"
                          f" generate_avro_schemas management command to save the signal schema.")
            data_dict = generate_test_data_for_schema(old_schema)

            # write to bytes using stored schema
            stored_out = io.BytesIO()
            schemaless_writer(stored_out, old_schema, data_dict)
            stored_out.seek(0)
            stored_event_bytes = stored_out.read()

            data_file_stored = io.BytesIO(stored_event_bytes)

            # read bytes using current schema
            schemaless_reader(data_file_stored, reader_schema=schema_dict, writer_schema=old_schema)

    def test_full_serialize_deserialize(self):
        SIGNAL = create_simple_signal({"test_data": EventData})
        event_data = {"test_data": EventData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )}
        serialized = serialize_event_data_to_bytes(event_data, SIGNAL)
        deserialized = deserialize_bytes_to_event_data(serialized, SIGNAL)
        self.assertIsInstance(deserialized["test_data"], EventData)
        self.assertEqual(deserialized, event_data)
        # ensure signal can actually send deserialized event data
        SIGNAL.send_event(**deserialized)

    def test_full_serialize_deserialize_with_optional_fields(self):
        SIGNAL = create_simple_signal({"test_data": NestedAttrsWithDefaults})
        event_data = {"test_data": NestedAttrsWithDefaults(field_0=SimpleAttrsWithDefaults())}
        serialized = serialize_event_data_to_bytes(event_data, SIGNAL)
        deserialized = deserialize_bytes_to_event_data(serialized, SIGNAL)
        self.assertIsInstance(deserialized["test_data"], NestedAttrsWithDefaults)
        self.assertEqual(deserialized, event_data)
        # ensure signal can actually send deserialized event data
        SIGNAL.send_event(**deserialized)



"""Tests for avro.serializer module."""

import json
from datetime import datetime

import pytest
from django.test import TestCase
from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2, LibraryUsageLocatorV2

from openedx_events.event_bus.avro.serializer import AvroSignalSerializer, serialize_event_data_to_bytes
from openedx_events.event_bus.avro.tests.test_utilities import (
    CustomAttrsWithDefaults,
    CustomAttrsWithoutDefaults,
    EventData,
    NestedAttrsWithDefaults,
    NestedNonAttrs,
    NonAttrs,
    SimpleAttrs,
    SimpleAttrsWithDefaults,
    SpecialSerializer,
    SubTestData0,
    SubTestData1,
    create_simple_signal,
)
from openedx_events.tests.utils import FreezeSignalCacheMixin


class TestAvroSignalSerializerCache(FreezeSignalCacheMixin, TestCase):
    """Tests for AvroSignalSerializer."""

    def test_schema_string(self):
        """
        Test JSON round-trip; schema creation is tested more fully in test_schema.py.
        """
        SIGNAL = create_simple_signal({
            "data": SimpleAttrs
        })
        actual_schema = json.loads(AvroSignalSerializer(SIGNAL).schema_string())
        expected_schema = {
            'name': 'CloudEvent',
            'type': 'record',
            'doc': 'Avro Event Format for CloudEvents created with openedx_events/schema',
            'namespace': 'simple.signal',
            'fields': [
                {
                    'name': 'data',
                    'type': {
                        'name': 'SimpleAttrs',
                        'type': 'record',
                        'fields': [
                            {'name': 'boolean_field', 'type': 'boolean'},
                            {'name': 'int_field', 'type': 'long'},
                            {'name': 'float_field', 'type': 'double'},
                            {'name': 'bytes_field', 'type': 'bytes'},
                            {'name': 'string_field', 'type': 'string'},
                        ]
                    }
                }
            ]
        }
        assert actual_schema == expected_schema

    def test_convert_event_data_to_dict(self):
        """
        Tests that an event with complex attrs objects can be converted to a dict.
        """

        # A test record that we can try to serialize to avro.
        test_data = EventData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )
        SIGNAL = create_simple_signal({"test_data": EventData})

        serializer = AvroSignalSerializer(SIGNAL)

        data_dict = serializer.to_dict({"test_data": test_data})
        expected_dict = {
            "test_data": {
                "course_id": "bar.course",
                "sub_name": "foo",
                "sub_test_0": {"course_id": "a.nother.course", "sub_name": "a.sub.name"},
                "sub_test_1": {
                    "course_id": "b.uber.another.course",
                    "sub_name": "b.uber.sub.name",
                },
            }
        }
        self.assertDictEqual(data_dict, expected_dict)

    def test_default_datetime_serialization(self):
        SIGNAL = create_simple_signal({"birthday": datetime})
        serializer = AvroSignalSerializer(SIGNAL)
        birthday = datetime(year=1989, month=9, day=6)
        test_data = {"birthday": birthday}
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"birthday": birthday.isoformat()})

    def test_default_coursekey_serialization(self):
        SIGNAL = create_simple_signal({"course": CourseKey})
        serializer = AvroSignalSerializer(SIGNAL)
        course_key = CourseKey.from_string("course-v1:edX+DemoX.1+2014")
        test_data = {"course": course_key}
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"course": str(course_key)})

    def test_default_usagekey_serialization(self):
        """
        Test serialization of UsageKey
        """
        SIGNAL = create_simple_signal({"usage_key": UsageKey})
        serializer = AvroSignalSerializer(SIGNAL)
        usage_key = UsageKey.from_string(
            "block-v1:edx+DemoX+Demo_course+type@video+block@UaEBjyMjcLW65gaTXggB93WmvoxGAJa0JeHRrDThk",
        )
        test_data = {"usage_key": usage_key}
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"usage_key": str(usage_key)})

    def test_default_librarylocatorv2_serialization(self):
        """
        Test serialization of LibraryLocatorV2
        """
        SIGNAL = create_simple_signal({"library_key": LibraryLocatorV2})
        serializer = AvroSignalSerializer(SIGNAL)
        library_key = LibraryLocatorV2.from_string("lib:MITx:reallyhardproblems")
        test_data = {"library_key": library_key}
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"library_key": str(library_key)})

    def test_default_libraryusagelocatorv2_serialization(self):
        """
        Test serialization of LibraryUsageLocatorV2
        """
        SIGNAL = create_simple_signal({"usage_key": LibraryUsageLocatorV2})
        serializer = AvroSignalSerializer(SIGNAL)
        usage_key = LibraryUsageLocatorV2.from_string("lb:MITx:reallyhardproblems:problem:problem1")
        test_data = {"usage_key": usage_key}
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"usage_key": str(usage_key)})

    def test_serialization_with_custom_serializer(self):
        SIGNAL = create_simple_signal({"test_data": NonAttrs})

        serializer = SpecialSerializer(SIGNAL)
        test_data = {
            "test_data": NonAttrs("a.val", "a.nother.val")
        }
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"test_data": "a.val:a.nother.val"})

    def test_serialization_with_custom_serializer_on_nested_fields(self):
        SIGNAL = create_simple_signal({"test_data": NestedNonAttrs})
        test_data = {
            "test_data": NestedNonAttrs(field_0=NonAttrs("a.val", "a.nother.val"))
        }
        serializer = SpecialSerializer(SIGNAL)
        test_data = {"test_data": NestedNonAttrs(field_0=NonAttrs("a.val", "a.nother.val"))}
        data_dict = serializer.to_dict(test_data)
        self.assertDictEqual(data_dict, {"test_data": {
            "field_0": "a.val:a.nother.val"
        }})

    def test_serialization_of_optional_simple_fields(self):
        SIGNAL = create_simple_signal({
            "data": SimpleAttrsWithDefaults
        })
        serializer = AvroSignalSerializer(SIGNAL)
        event_data = {"data": SimpleAttrsWithDefaults()}
        data_dict = serializer.to_dict(event_data)
        self.assertDictEqual(data_dict, {"data": {'boolean_field': None,
                                                  'bytes_field': None,
                                                  'float_field': None,
                                                  'int_field': None,
                                                  'string_field': None,
                                                  'attrs_field': None}})

    def test_serialization_of_optional_custom_fields(self):
        SIGNAL = create_simple_signal({"data": CustomAttrsWithDefaults})
        serializer = AvroSignalSerializer(SIGNAL)
        event_data = {"data": CustomAttrsWithDefaults(coursekey_field=None, datetime_field=None)}
        data_dict = serializer.to_dict(event_data)
        self.assertDictEqual(data_dict, {"data": {'coursekey_field': None,
                                                  'datetime_field': None}})

    def test_serialization_of_none_mandatory_custom_fields(self):
        """Check that None isn't accepted if field not optional."""
        SIGNAL = create_simple_signal({"data": CustomAttrsWithoutDefaults})
        serializer = AvroSignalSerializer(SIGNAL)
        event_data = {"data": CustomAttrsWithoutDefaults(coursekey_field=None, datetime_field=None)}
        with pytest.raises(Exception) as excinfo:
            serializer.to_dict(event_data)
        assert excinfo.value.args[0] == "None cannot be handled by custom serializers (and default=None was not set)"

    def test_serialization_of_nested_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": NestedAttrsWithDefaults
        })
        serializer = AvroSignalSerializer(SIGNAL)

        event_data = {"data": NestedAttrsWithDefaults(field_0=SimpleAttrsWithDefaults())}
        data_dict = serializer.to_dict(event_data)
        self.assertDictEqual(data_dict, {"data": {"field_0": {'boolean_field': None,
                                                              'bytes_field': None,
                                                              'float_field': None,
                                                              'int_field': None,
                                                              'string_field': None,
                                                              'attrs_field': None
                                                              }}})

    def test_serialize_event_data_to_bytes(self):
        """
        Test serialize_event_data_to_bytes utility function.
        """
        SIGNAL = create_simple_signal({"test_data": EventData})
        event_data = {"test_data": EventData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )}
        serialized = serialize_event_data_to_bytes(event_data, SIGNAL)
        expected = b'\x06foo\x14bar.course\x14a.sub.name\x1ea.nother.course\x1eb.uber.sub.name*b.uber.another.course'
        self.assertEqual(serialized, expected)



"""
Tests for event_bus.avro.schema module
"""
from typing import List
from unittest import TestCase

from openedx_events.event_bus.avro.schema import schema_from_signal
from openedx_events.event_bus.avro.tests.test_utilities import (
    EventData,
    NestedAttrsWithDefaults,
    NestedNonAttrs,
    NonAttrs,
    SimpleAttrs,
    SimpleAttrsWithDefaults,
    SubTestData0,
    SubTestData1,
    create_simple_signal,
)


class TestSchemaGeneration(TestCase):
    """
    Test Avro schema generation.
    """

    def setUp(self) -> None:
        super().setUp()
        self.maxDiff = None

    def test_simple_schema_generation(self):
        SIGNAL = create_simple_signal({"event_data": SimpleAttrs})
        schema = schema_from_signal(SIGNAL)
        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
            'namespace': 'simple.signal',
            "fields": [
                {"name": "event_data", "type":
                    {"name": "SimpleAttrs", "type": "record", "fields": [
                        {"name": "boolean_field", "type": "boolean"},
                        {"name": "int_field", "type": "long"},
                        {"name": "float_field", "type": "double"},
                        {"name": "bytes_field", "type": "bytes"},
                        {"name": "string_field", "type": "string"},
                    ]
                     }
                 }
            ],
        }

        self.assertDictEqual(schema, expected_dict)

    def test_nested_attrs_object_serialization(self):
        SIGNAL = create_simple_signal({"test_data": EventData})
        schema = schema_from_signal(SIGNAL)

        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
            'namespace': 'simple.signal',
            "fields": [
                {"name": "test_data", "type":
                    {"name": "EventData", "type": "record", "fields": [
                        {"name": "sub_name", "type": "string"},
                        {"name": "course_id", "type": "string"},
                        {"name": "sub_test_0", "type": {
                            "name": "SubTestData0",
                            "type": "record",
                            "fields": [
                                {"name": "sub_name", "type": "string"},
                                {"name": "course_id", "type": "string"},
                            ]
                        }},
                        {"name": "sub_test_1", "type": {
                            "name": "SubTestData1",
                            "type": "record",
                            "fields": [
                                {"name": "sub_name", "type": "string"},
                                {"name": "course_id", "type": "string"},
                            ]
                        }},
                    ]}
                 },
            ],
        }

        self.assertDictEqual(schema, expected_dict)

    def test_multiple_top_level_fields(self):
        SIGNAL = create_simple_signal({
            "top_level_key_0": SubTestData0,
            "top_level_key_1": SubTestData1,
        })
        schema = schema_from_signal(SIGNAL)

        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
            'namespace': 'simple.signal',
            "fields": [
                {"name": "top_level_key_0", "type":
                    {
                        "name": "SubTestData0",
                        "type": "record",
                        "fields": [
                            {"name": "sub_name", "type": "string"},
                            {"name": "course_id", "type": "string"},
                        ]
                    }
                 },
                {"name": "top_level_key_1", "type":
                    {
                        "name": "SubTestData1",
                        "type": "record",
                        "fields": [
                            {"name": "sub_name", "type": "string"},
                            {"name": "course_id", "type": "string"},
                        ]
                    }
                 },
            ],
        }
        self.assertDictEqual(schema, expected_dict)

    def test_schema_for_custom_type(self):
        SIGNAL = create_simple_signal({"test_data": NonAttrs})
        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
            'namespace': 'simple.signal',
            "fields": [
                {"name": "test_data", "type": "string"}
            ],
        }
        schema = schema_from_signal(SIGNAL, custom_type_to_avro_type={NonAttrs: "string"})
        self.assertDictEqual(schema, expected_dict)

    def test_schema_for_nested_custom_type(self):
        SIGNAL = create_simple_signal({"test_data": NestedNonAttrs})
        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
            'namespace': 'simple.signal',
            "fields": [
                {"name": "test_data", "type": {
                    "name": "NestedNonAttrs",
                    "type": "record",
                    "fields": [
                        {'name': 'field_0', 'type': 'string'}
                    ]
                }}
            ],
        }
        schema = schema_from_signal(SIGNAL, custom_type_to_avro_type={NonAttrs: "string"})
        self.assertDictEqual(schema, expected_dict)

    def test_schema_for_types_with_defaults(self):
        SIGNAL = create_simple_signal({"test_data": SimpleAttrsWithDefaults})
        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
            'namespace': 'simple.signal',
            "fields": [
                {"name": "test_data", "type":
                    {"name": "SimpleAttrsWithDefaults", "type": "record", "fields": [
                        {"name": "boolean_field", "type": ["null", "boolean"], "default": None},
                        {"name": "int_field", "type": ["null", "long"], "default": None},
                        {"name": "float_field", "type": ["null", "double"], "default": None},
                        {"name": "bytes_field", "type": ["null", "bytes"], "default": None},
                        {"name": "string_field", "type": ["null", "string"], "default": None},
                        {"name": "attrs_field", "default": None,
                         "type": ["null",
                                  {"name": "SimpleAttrs", "type": "record", "fields": [
                                      {"name": "boolean_field", "type": "boolean"},
                                      {"name": "int_field", "type": "long"},
                                      {"name": "float_field", "type": "double"},
                                      {"name": "bytes_field", "type": "bytes"},
                                      {"name": "string_field", "type": "string"},
                                  ]
                                   }]}
                    ]}
                 }
            ],
        }
        schema = schema_from_signal(SIGNAL)
        self.assertDictEqual(schema, expected_dict)

    def test_schema_for_types_with_nested_defaults(self):
        SIGNAL = create_simple_signal({"test_data": NestedAttrsWithDefaults})
        expected_dict = {
            'name': 'CloudEvent',
            'type': 'record',
            'doc': 'Avro Event Format for CloudEvents created with openedx_events/schema',
            'namespace': 'simple.signal',
            'fields': [{
                'name': 'test_data',
                'type': {
                    'name': 'NestedAttrsWithDefaults',
                    'type': 'record',
                    'fields': [{
                        'name': 'field_0',
                        'type': {
                            'name': 'SimpleAttrsWithDefaults',
                            'type': 'record',
                            'fields': [
                                {'name': 'boolean_field', 'type': ['null', 'boolean'], 'default': None},
                                {'name': 'int_field', 'type': ['null', 'long'], 'default': None},
                                {'name': 'float_field', 'type': ['null', 'double'], 'default': None},
                                {'name': 'bytes_field', 'type': ['null', 'bytes'], 'default': None},
                                {'name': 'string_field', 'type': ['null', 'string'], 'default': None},
                                {'name': 'attrs_field', 'default': None,
                                 'type': ['null', {
                                     'name': 'SimpleAttrs',
                                     'type': 'record',
                                     'fields': [
                                         {'name': 'boolean_field', 'type': 'boolean'},
                                         {'name': 'int_field', 'type': 'long'},
                                         {'name': 'float_field', 'type': 'double'},
                                         {'name': 'bytes_field', 'type': 'bytes'},
                                         {'name': 'string_field', 'type': 'string'
                                          }]
                                 }],
                                 }]
                        }
                    }]
                }
            }]
        }
        schema = schema_from_signal(SIGNAL, custom_type_to_avro_type={NonAttrs: "string"})
        self.assertDictEqual(schema, expected_dict)

    def test_throw_exception_on_unextended_custom_type(self):
        class UnextendedClass:
            pass

        SIGNAL = create_simple_signal({"unextended_class": UnextendedClass})
        with self.assertRaises(TypeError):
            schema_from_signal(SIGNAL)

    def test_throw_exception_to_list_or_dict_types_without_annotation(self):
        LIST_SIGNAL = create_simple_signal({"list_input": list})
        DICT_SIGNAL = create_simple_signal({"list_input": dict})
        LIST_WITHOUT_ANNOTATION_SIGNAL = create_simple_signal({"list_input": List})
        with self.assertRaises(Exception):
            schema_from_signal(LIST_SIGNAL)

        with self.assertRaises(Exception):
            schema_from_signal(DICT_SIGNAL)

        with self.assertRaises(TypeError):
            schema_from_signal(LIST_WITHOUT_ANNOTATION_SIGNAL)

    def test_list_with_annotation_works(self):
        LIST_SIGNAL = create_simple_signal({"list_input": List[int]})
        expected_dict = {
            'name': 'CloudEvent',
            'type': 'record',
            'doc': 'Avro Event Format for CloudEvents created with openedx_events/schema',
            'namespace': 'simple.signal',
            'fields': [{
                'name': 'list_input',
                'type': {'type': 'array', 'items': 'long'},
            }],
        }
        schema = schema_from_signal(LIST_SIGNAL)
        self.assertDictEqual(schema, expected_dict)



"""
Standardized signals definitions for events within the architecture subdomain ``content_authoring``.

All signals defined in this module must follow the name and versioning
conventions specified in OEP-41.

They also must comply with the payload definition specified in
docs/decisions/0003-events-payload.rst
"""
from openedx_events.content_authoring.data import (
    CertificateConfigData,
    ContentLibraryData,
    CourseCatalogData,
    CourseData,
    DuplicatedXBlockData,
    LibraryBlockData,
    XBlockData,
)
from openedx_events.tooling import OpenEdxPublicSignal

# .. event_type: org.openedx.content_authoring.course.catalog_info.changed.v1
# .. event_name: COURSE_CATALOG_INFO_CHANGED
# .. event_description: Fired when a course changes in Studio in a way that is relevant for catalog consumers.
# .. event_data: CourseCatalogData
COURSE_CATALOG_INFO_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.course.catalog_info.changed.v1",
    data={
        "catalog_info": CourseCatalogData,
    }
)

# .. event_type: org.openedx.content_authoring.xblock.created.v1
# .. event_name: XBLOCK_CREATED
# .. event_description: Fired when an XBlock is created.
# .. event_data: XBlockData
XBLOCK_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.xblock.created.v1",
    data={
        "xblock_info": XBlockData,
    }
)

# .. event_type: org.openedx.content_authoring.xblock.updated.v1
# .. event_name: XBLOCK_UPDATED
# .. event_description: Fired when an XBlock is updated.
# .. event_data: XBlockData
XBLOCK_UPDATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.xblock.updated.v1",
    data={
        "xblock_info": XBlockData,
    }
)

# .. event_type: org.openedx.content_authoring.xblock.published.v1
# .. event_name: XBLOCK_PUBLISHED
# .. event_description: Fired when an XBlock is published. If a parent block
#       with changes in one or more child blocks is published, only a single
#       XBLOCK_PUBLISHED event is fired with parent block details.
#       For example: If a section is published with changes in multiple units,
#       only a single event is fired with section details like :
#       `XBlockData(usage_key="section-usage-key", block_type="chapter")`
# .. event_data: XBlockData
XBLOCK_PUBLISHED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.xblock.published.v1",
    data={
        "xblock_info": XBlockData,
    }
)


# .. event_type: org.openedx.content_authoring.xblock.deleted.v1
# .. event_name: XBLOCK_DELETED
# .. event_description: Fired when an XBlock is deleted.
# .. event_data: XBlockData
XBLOCK_DELETED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.xblock.deleted.v1",
    data={
        "xblock_info": XBlockData,
    }
)


# .. event_type: org.openedx.content_authoring.xblock.duplicated.v1
# .. event_name: XBLOCK_DUPLICATED
# .. event_description: Fired when an XBlock is duplicated in Studio.
# .. event_data: DuplicatedXBlockData
XBLOCK_DUPLICATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.xblock.duplicated.v1",
    data={
        "xblock_info": DuplicatedXBlockData,
    }
)


# .. event_type: org.openedx.content_authoring.course.certificate_config.changed.v1
# .. event_name: COURSE_CERTIFICATE_CONFIG_CHANGED
# .. event_description: Fired when a course certificate configuration changes in Studio.
#      Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
# .. event_data: CertificateConfigData
COURSE_CERTIFICATE_CONFIG_CHANGED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.course.certificate_config.changed.v1",
    data={
        "certificate_config": CertificateConfigData,
    }
)

# .. event_type: org.openedx.content_authoring.course.certificate_config.deleted.v1
# .. event_name: COURSE_CERTIFICATE_CONFIG_DELETED
# .. event_description: Fired when a course certificate configuration deletes in Studio.
#      Warning: This event is currently incompatible with the event bus, list/dict cannot be serialized yet
# .. event_data: CertificateConfigData
COURSE_CERTIFICATE_CONFIG_DELETED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.course.certificate_config.deleted.v1",
    data={
        "certificate_config": CertificateConfigData,
    }
)

# .. event_type: org.openedx.content_authoring.course.created.v1
# .. event_name: COURSE_CREATED
# .. event_description: emitted when a course is created
# .. event_data: CourseData
COURSE_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.course.created.v1",
    data={
        "course": CourseData,
    }
)

# .. event_type: org.openedx.content_authoring.content_library.created.v1
# .. event_name: CONTENT_LIBRARY_CREATED
# .. event_description: emitted when a content library is created
# .. event_data: ContentLibraryData
CONTENT_LIBRARY_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.content_library.created.v1",
    data={
        "content_library": ContentLibraryData,
    }
)

# .. event_type: org.openedx.content_authoring.content_library.updated.v1
# .. event_name: CONTENT_LIBRARY_UPDATED
# .. event_description: emitted when a content library is updated
# .. event_data: ContentLibraryData
CONTENT_LIBRARY_UPDATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.content_library.updated.v1",
    data={
        "content_library": ContentLibraryData,
    }
)

# .. event_type: org.openedx.content_authoring.content_library.deleted.v1
# .. event_name: CONTENT_LIBRARY_DELETED
# .. event_description: emitted when a content library is deleted
# .. event_data: ContentLibraryData
CONTENT_LIBRARY_DELETED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.content_library.deleted.v1",
    data={
        "content_library": ContentLibraryData,
    }
)

# .. event_type: org.openedx.content_authoring.library_block.created.v1
# .. event_name: LIBRARY_BLOCK_CREATED
# .. event_description: emitted when a library block is created
# .. event_data: LibraryBlockData
LIBRARY_BLOCK_CREATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.library_block.created.v1",
    data={
        "library_block": LibraryBlockData,
    }
)

# .. event_type: org.openedx.content_authoring.library_block.updated.v1
# .. event_name: LIBRARY_BLOCK_UPDATED
# .. event_description: emitted when a library block is updated
# .. event_data: LibraryBlockData
LIBRARY_BLOCK_UPDATED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.library_block.updated.v1",
    data={
        "library_block": LibraryBlockData,
    }
)

# .. event_type: org.openedx.content_authoring.library_block.deleted.v1
# .. event_name: LIBRARY_BLOCK_DELETED
# .. event_description: emitted when a library block is deleted
# .. event_data: LibraryBlockData
LIBRARY_BLOCK_DELETED = OpenEdxPublicSignal(
    event_type="org.openedx.content_authoring.library_block.deleted.v1",
    data={
        "library_block": LibraryBlockData,
    }
)



"""
openedx_events Django application initialization.
"""

from django.apps import AppConfig
from django.conf import settings

from openedx_events.event_bus import get_producer
from openedx_events.exceptions import ProducerConfigurationError
from openedx_events.tooling import OpenEdxPublicSignal, load_all_signals


def general_signal_handler(sender, signal, **kwargs):  # pylint: disable=unused-argument
    """
    Signal handler for producing events to configured event bus.
    """
    event_type_producer_configs = getattr(settings, "EVENT_BUS_PRODUCER_CONFIG", {}).get(signal.event_type, {})
    # event_type_producer_configs should look something like
    # {
    #        "topic_a": { "event_key_field": "my.key.field", "enabled": True },
    #        "topic_b": { "event_key_field": "my.key.field", "enabled": False }
    # }"
    event_data = {key: kwargs.get(key) for key in signal.init_data}

    for topic in event_type_producer_configs.keys():
        if event_type_producer_configs[topic]["enabled"] is True:
            get_producer().send(
                signal=signal,
                topic=topic,
                event_key_field=event_type_producer_configs[topic]["event_key_field"],
                event_data=event_data,
                event_metadata=kwargs["metadata"],
            )


class OpenedxEventsConfig(AppConfig):
    """
    Configuration for the openedx_events Django application.
    """

    name = "openedx_events"

    def _get_validated_signal_config(self, event_type, configuration):
        """
        Validate signal configuration format.

        Example expected signal configuration:
        {
            "topic_a": { "event_key_field": "my.key.field", "enabled": True },
            "topic_b": { "event_key_field": "my.key.field", "enabled": False }
        }

        Raises:
            ProducerConfigurationError: If configuration is not valid.
        """
        if not isinstance(configuration, dict):
            raise ProducerConfigurationError(
                event_type=event_type,
                message="Configuration for event_types should be a dict"
            )
        try:
            signal = OpenEdxPublicSignal.get_signal_by_type(event_type)
        except KeyError as exc:
            raise ProducerConfigurationError(message=f"No OpenEdxPublicSignal of type: '{event_type}'.") from exc
        for _, topic_configuration in configuration.items():
            if not isinstance(topic_configuration, dict):
                raise ProducerConfigurationError(
                    event_type=event_type,
                    message="One of the configuration objects is not a dictionary"
                )
            expected_keys = {"event_key_field": str, "enabled": bool}
            for expected_key, expected_type in expected_keys.items():
                if expected_key not in topic_configuration.keys():
                    raise ProducerConfigurationError(
                        event_type=event_type,
                        message=f"One of the configuration object is missing '{expected_key}' key."
                    )
                if not isinstance(topic_configuration[expected_key], expected_type):
                    raise ProducerConfigurationError(
                        event_type=event_type,
                        message=(f"Expected type: {expected_type} for '{expected_key}', "
                                 f"found: {type(topic_configuration[expected_key])}")
                    )
        return signal

    def ready(self):
        """
        Read `EVENT_BUS_PRODUCER_CONFIG` setting and connects appropriate handlers to the events based on it.

        Example expected configuration:
        {
            "org.openedx.content_authoring.xblock.deleted.v1" : {
                "topic_a": { "event_key_field": "xblock_info.usage_key", "enabled": True },
                "topic_b": { "event_key_field": "xblock_info.usage_key", "enabled": False }
            },
            "org.openedx.content_authoring.course.catalog_info.changed.v1" : {
                "topic_c": {"event_key_field": "course_info.course_key", "enabled": True }
            }
        }

        Raises:
            ProducerConfigurationError: If `EVENT_BUS_PRODUCER_CONFIG` is not valid.
        """
        load_all_signals()
        signals_config = getattr(settings, "EVENT_BUS_PRODUCER_CONFIG", {})
        if not isinstance(signals_config, dict):
            raise ProducerConfigurationError(
                message=("Setting 'EVENT_BUS_PRODUCER_CONFIG' should be a dictionary with event_type as"
                         " key and list or tuple of config dictionaries as values")
            )
        for event_type, configurations in signals_config.items():
            signal = self._get_validated_signal_config(event_type, configurations)
            signal.connect(general_signal_handler)
        return super().ready()



"""
Data attributes for events within the architecture subdomain ``content_authoring``.

These attributes follow the form of attr objects specified in OEP-49 data
pattern.

The attributes for the events come from the CourseDetailView in the LMS, with some unused fields removed
(see deprecation proposal at https://github.com/openedx/public-engineering/issues/160)
"""
from datetime import datetime
from typing import BinaryIO, List

import attr
from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2, LibraryUsageLocatorV2


@attr.s(frozen=True)
class CourseData:
    """
    Attributes defined for Open edX Course object.

    Arguments:
        course_key (CourseKey): identifier of the Course object.
    """

    course_key = attr.ib(type=CourseKey)


@attr.s(frozen=True)
class CourseScheduleData:
    """
    Data describing course scheduling.

    Arguments:
        start (datetime): course start date
        pacing (str): 'instructor' or 'self'
        end (datetime): course end date (optional)
        enrollment_start (datetime): start of course enrollment (optional)
        enrollment_end (datetime): end of course enrollment (optional)
    """

    start = attr.ib(type=datetime)
    pacing = attr.ib(type=str)
    end = attr.ib(type=datetime, default=None)
    enrollment_start = attr.ib(type=datetime, default=None)
    enrollment_end = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class CourseCatalogData:
    """
    Data needed for a course catalog entry.

    Arguments:
        course_key (CourseKey): identifier of the Course object.
        name (str): course name
        schedule_data (CourseScheduleData): scheduling information for the course
        hidden (bool): whether the course is hidden from search (optional)
        invitation_only (bool): whether the course requires an invitation to enroll
    """

    # basic identifiers
    course_key = attr.ib(type=CourseKey)
    name = attr.ib(type=str)

    # additional marketing information
    schedule_data = attr.ib(type=CourseScheduleData)
    hidden = attr.ib(type=bool, default=False)
    invitation_only = attr.ib(type=bool, default=False)


@attr.s(frozen=True)
class XBlockData:
    """
    Data about changed XBlock.

    Arguments:
        usage_key (UsageKey): identifier of the XBlock object.
        block_type (str): type of block.
        version (UsageKey): identifier of the XBlock object with branch and version data (optional). This
        could be used to get the exact version of the XBlock object.
    """

    usage_key = attr.ib(type=UsageKey)
    block_type = attr.ib(type=str)
    version = attr.ib(type=UsageKey, default=None, kw_only=True)


@attr.s(frozen=True)
class DuplicatedXBlockData(XBlockData):
    """
    Data about duplicated XBlock.

    This class extends XBlockData to include source_usage_key.

    Arguments:
        source_usage_key (UsageKey): identifier of the source XBlock object.
    """

    source_usage_key = attr.ib(type=UsageKey)


@attr.s(frozen=True)
class CertificateSignatoryData:
    """
    Attributes defined for Open edX CertificateSignatory data object.

    Arguments:
        image (BinaryIO): certificate signature image.
        name (str): name of signatory.
        organization (str): organization that signatory belongs to.
        title (int): signatory title.
    """

    # Note: Please take care that the image field is BinaryIO, which means
    # that a file can be passed as an array of bytes. Watch the size of this file.
    # It can potentially be large, making it difficult to pass this data structure through the Event Bus
    # (CloudEvent messages, which should be 64K or less) and store it on disk space.
    # We suggest referring to MAX_ASSET_UPLOAD_FILE_SIZE_IN_MB, i.e. restriction in the Studio for such cases.
    image = attr.ib(type=BinaryIO)
    # end Note
    name = attr.ib(type=str)
    organization = attr.ib(type=str)
    title = attr.ib(type=str)


@attr.s(frozen=True)
class CertificateConfigData:
    """
    Attributes defined for Open edX CertificateConfig data object.

    Arguments:
        certificate_type (str): certificate type. Possible types are certificate relevant course modes:
         - credit,
         - verified,
         - professional,
         - no-id-professional,
         - executive-education,
         - paid-executive-education,
         - paid-bootcamp,
         - masters.
        course_key (CourseKey): identifier of the Course object.
        title (str): certificate title.
        signatories (List[CertificateSignatoryData]): contains a collection of signatures
        that belong to the certificate configuration.
        is_active (bool): indicates whether the certifivate configuration is active.
    """

    certificate_type = attr.ib(type=str)
    course_key = attr.ib(type=CourseKey)
    title = attr.ib(type=str)
    signatories = attr.ib(type=List[CertificateSignatoryData], factory=list)
    is_active = attr.ib(type=bool, default=False)


@attr.s(frozen=True)
class ContentLibraryData:
    """
    Data about changed ContentLibrary.

    Arguments:
        library_key (LibraryLocatorV2): a key that represents a Blockstore-based content library.
        update_blocks (bool): flag that indicates whether the content library blocks indexes should be updated
    """

    library_key = attr.ib(type=LibraryLocatorV2)
    update_blocks = attr.ib(type=bool, default=False)


@attr.s(frozen=True)
class LibraryBlockData:
    """
    Data about changed LibraryBlock.

    Arguments:
        library_key (LibraryLocatorV2): a key that represents a Blockstore-based content library.
        usage_key (LibraryUsageLocatorV2): a key that represents a XBlock in a Blockstore-based content library.
    """

    library_key = attr.ib(type=LibraryLocatorV2)
    usage_key = attr.ib(type=LibraryUsageLocatorV2)



"""
This file contains all test for the tooling.py file.

Classes:
    EventsToolingTest: Test events tooling.
"""
import datetime
import sys
from contextlib import contextmanager
from unittest.mock import Mock, patch
from uuid import UUID, uuid1

import attr
import ddt
import pytest
from django.test import TestCase, override_settings

from openedx_events.data import EventsMetadata
from openedx_events.exceptions import SenderValidationError
from openedx_events.tests.utils import FreezeSignalCacheMixin
from openedx_events.tooling import OpenEdxPublicSignal, _process_all_signals_modules, load_all_signals


@contextmanager
def receivers_attached(signal, receivers):
    """
    Attach the receivers to the signal for the duration of the context.
    """
    try:
        for receiver in receivers:
            signal.connect(receiver)

        yield
    finally:
        for receiver in receivers:
            signal.disconnect(receiver)


@ddt.ddt
class OpenEdxPublicSignalTestCache(FreezeSignalCacheMixin, TestCase):
    """
    Test cases for Open edX events base class.
    """

    def setUp(self):
        """
        Setup common conditions for every test case.
        """
        super().setUp()
        self.event_type = "org.openedx.learning.session.login.completed.v1"
        self.user_mock = Mock()
        self.data_attr = {
            "user": Mock,
        }
        self.public_signal = OpenEdxPublicSignal(
            event_type=self.event_type,
            data=self.data_attr,
        )

        self.receiver_error = Exception("fake error")

        def error_receiver(*args, **kwargs):
            raise self.receiver_error

        self.ok_receiver = Mock(return_value="success")
        self.error_receiver = error_receiver

    def test_string_representation(self):
        """
        This methods checks the string representation for events base class.

        Expected behavior:
            The representation contains the event_type.
        """
        self.assertIn(self.event_type, str(self.public_signal))

    def test_get_signal_by_type(self):
        """
        Test found and not-found behavior.
        """
        assert isinstance(
            OpenEdxPublicSignal.get_signal_by_type('org.openedx.learning.session.login.completed.v1'),
            OpenEdxPublicSignal
        )

        with pytest.raises(KeyError):
            OpenEdxPublicSignal.get_signal_by_type('xxx')

    @override_settings(SERVICE_VARIANT="lms")
    @patch("openedx_events.data.openedx_events")
    @patch("openedx_events.data.socket")
    @patch("openedx_events.data.datetime")
    def test_generate_signal_metadata(self, datetime_mock, socket_mock, events_package_mock):
        """
        This methods tests getting the generated metadata for an event.

        Expected behavior:
            Returns the metadata containing information about the event.
        """
        events_package_mock.__version__ = "0.1.0"
        socket_mock.gethostname.return_value = "edx.devstack.lms"
        expected_time = datetime.datetime.now(datetime.timezone.utc)
        datetime_mock.now.return_value = expected_time
        expected_metadata = {
            "event_type": self.event_type,
            "minorversion": 0,
            "source": "openedx/lms/web",
            "sourcehost": "edx.devstack.lms",
            "sourcelib": [0, 1, 0],
            "time": expected_time,
        }

        metadata = self.public_signal.generate_signal_metadata()

        self.assertDictContainsSubset(expected_metadata, attr.asdict(metadata))
        self.assertIsInstance(metadata.id, UUID)

    @override_settings(SERVICE_VARIANT="lms")
    @patch("openedx_events.data.openedx_events")
    @patch("openedx_events.data.socket")
    def test_generate_signal_metadata_with_valid_time(self, socket_mock, events_package_mock):
        """
        Tests getting the generated metadata for an event, providing a valid time in UTC.

        Expected behavior:
            Returns the metadata containing information about the event.
        """
        events_package_mock.__version__ = "0.1.0"
        socket_mock.gethostname.return_value = "edx.devstack.lms"
        expected_time = datetime.datetime.now(datetime.timezone.utc)
        expected_metadata = {
            "event_type": self.event_type,
            "minorversion": 0,
            "source": "openedx/lms/web",
            "sourcehost": "edx.devstack.lms",
            "sourcelib": [0, 1, 0],
            "time": expected_time,
        }

        metadata = self.public_signal.generate_signal_metadata(time=expected_time)

        self.assertDictContainsSubset(expected_metadata, attr.asdict(metadata))
        self.assertIsInstance(metadata.id, UUID)

    @ddt.data(
        (1, TypeError, "'time' must be <class 'datetime.datetime'",),
        # WARNING: utcnow() has no timezone, and could be misinterpreted in local time
        (datetime.datetime.utcnow(), ValueError, "'time' must have timezone.utc",),
    )
    @ddt.unpack
    def test_generate_signal_metadata_fails_with_invalid_time(
        self, invalid_time, error_class, error_message
    ):
        """
        Tests getting generated metadata for an event fails with a non-UTC time.

        Expected behavior:
            Raises an exception
        """
        with self.assertRaisesMessage(error_class, error_message):
            self.public_signal.generate_signal_metadata(time=invalid_time)

    @patch("openedx_events.tooling.OpenEdxPublicSignal.generate_signal_metadata")
    @patch("openedx_events.tooling.Signal.send")
    def test_send_event_allow_failure_successfully(self, send_mock, fake_metadata):
        """
        This method tests the process of sending an event that's allowed to fail.

        Expected behavior:
            The event is sent as a django signal with send method.
        """
        expected_metadata = Mock(some_data="some_data")
        fake_metadata.return_value = expected_metadata
        self.public_signal.allow_send_event_failure()

        self.public_signal.send_event(user=self.user_mock)

        send_mock.assert_called_once_with(
            sender=None,
            user=self.user_mock,
            metadata=expected_metadata,
        )

    @patch("openedx_events.tooling.OpenEdxPublicSignal.generate_signal_metadata")
    @patch("openedx_events.tooling.log", autospec=True)
    @patch("openedx_events.tooling.format_responses", autospec=True, return_value="fake-output")
    def test_send_robust_event_successfully(self, format_responses_mock, log_mock, fake_metadata):
        """
        This method tests the process of sending an event that won't crash.

        Expected behavior:
            The event is sent as a django signal with send_robust method.
        """
        expected_metadata = Mock(some_data="some_data")
        fake_metadata.return_value = expected_metadata

        with receivers_attached(self.public_signal, [self.ok_receiver, self.error_receiver]):
            self.public_signal.send_event(user=self.user_mock)

        self.ok_receiver.assert_called_once_with(
            signal=self.public_signal, sender=None, user=self.user_mock, metadata=expected_metadata
        )
        # format_responses is mocked out because its output is
        # complicated enough to warrant its own set of tests.
        format_responses_mock.assert_called_once_with(
            [(self.ok_receiver, "success"), (self.error_receiver, self.receiver_error)], depth=2
        )
        log_mock.info.assert_called_once_with(
            "Responses of the Open edX Event <org.openedx.learning.session.login.completed.v1>: \nfake-output"
        )

    @patch("openedx_events.tooling.OpenEdxPublicSignal.generate_signal_metadata")
    def test_send_event_with_time(self, fake_metadata):
        """
        This method tests the process of sending an event with a time argument.

        Expected behavior:
            The generate_signal_metadata is called using the passed time.
        """
        expected_metadata = Mock(some_data="some_data")
        expected_time = datetime.datetime.now(datetime.timezone.utc)
        fake_metadata.return_value = expected_metadata

        self.public_signal.send_event(user=self.user_mock, time=expected_time)

        # generate_signal_metadata is fully tested elsewhere
        fake_metadata.assert_called_once_with(time=expected_time)

    @patch("openedx_events.tooling.OpenEdxPublicSignal._send_event_with_metadata")
    def test_send_event_with_custom_metadata(self, mock_send_event_with_metadata):
        """
        This method tests the process of sending an event with custom metadata.

        Expected behavior:
            The _send_event_with_metadata call is passed the appropriate metadata.

        Note:
            The _send_event_with_metadata is fully tested with the various send_event tests.
        """
        metadata = EventsMetadata(
            id=uuid1(),
            event_type=self.event_type,
            minorversion=99,
            source="mock-source",
            sourcehost="mock-sourcehost",
            time=datetime.datetime.now(datetime.timezone.utc),
            sourcelib=(6, 1, 7),
        )
        expected_response = "mock-response"
        mock_send_event_with_metadata.return_value = expected_response

        response = self.public_signal.send_event_with_custom_metadata(metadata, foo="bar")

        assert response == expected_response
        mock_send_event_with_metadata.assert_called_once_with(
            metadata=metadata, send_robust=True, foo="bar",
        )

    @ddt.data(
        (
            {"student": Mock()},
            "SenderValidationError org.openedx.learning.session.login.completed.v1: "
            "Missing required argument 'user'",
        ),
        (
            {"user": {"student": Mock()}},
            "SenderValidationError org.openedx.learning.session.login.completed.v1: "
            "The argument 'user' is not instance of the Class Attribute 'type'",
        ),
        (
            {"student": Mock(), "user": Mock()},
            "SenderValidationError org.openedx.learning.session.login.completed.v1: "
            "There's a mismatch between initialization data and send_event arguments",
        ),
    )
    @ddt.unpack
    def test_invalid_sender(self, send_arguments, exception_message):
        """
        This method tests sending an event with invalid setup on the sender
        side.

        Expected behavior:
            A SenderValidationError exception is raised.
        """
        with self.assertRaisesMessage(SenderValidationError, exception_message):
            self.public_signal.send_event(**send_arguments)

    def test_send_event_with_django(self):
        """
        This method tests sending an event using the `send` built-in Django
        method.

        Expected behavior:
            A warning is showed advising to use Open edX events custom
            send_signal method.
        """
        message = "Please, use 'send_event' when triggering an Open edX event."

        with self.assertWarns(Warning, msg=message):
            self.public_signal.send(sender=Mock())

    def test_send_robust_event_with_django(self):
        """
        This method tests sending an event using the `send` built-in Django
        method.

        Expected behavior:
            A warning is showed advicing to use Open edX events custom
            send_signal method.
        """
        message = "Please, use 'send_event' with send_robust equals to True when triggering an Open edX event."

        with self.assertWarns(Warning, msg=message):
            self.public_signal.send_robust(sender=Mock())

    @patch("openedx_events.tooling.Signal.send")
    def test_send_event_disabled(self, send_mock):
        """
        This method tests sending an event that has been disabled.

        Expected behavior:
            The Django Signal associated to the event is not sent.
        """
        self.public_signal.disable()

        result = self.public_signal.send_event(sender=Mock())

        send_mock.assert_not_called()
        self.assertListEqual([], result)


class TestLoadAllSignals(FreezeSignalCacheMixin, TestCase):
    """ Tests for the load_all_signals method"""
    def setUp(self):
        # load_all_signals does spooky things with module loading,
        # so save the current state of any loaded signals modules to avoid disrupting other tests
        super().setUp()
        self.old_signal_modules = {}

        def save_module(module_name):
            if module_name in sys.modules:
                self.old_signal_modules[module_name] = sys.modules[module_name]
        _process_all_signals_modules(save_module)

    def tearDown(self):
        for k, v in self.old_signal_modules.items():
            sys.modules[k] = v
        super().tearDown()

    def test_load_all_signals(self):
        """
        Tests load_all_signals loads all the signals in the entire library

        It's not the most robust since it just tests a few arbitrary signals but actually testing all the signals
        would require updating this test every time a new signal is added
        """

        # remove any existing imports of signals modules
        for k in self.old_signal_modules:
            sys.modules.pop(k)

        # this class uses FreezeSignalCacheMixin so we can safely remove everything from the OpenEdxPublicSignal
        # cache and it shouldn't affect any other tests
        OpenEdxPublicSignal._mapping = {}  # pylint: disable=protected-access
        OpenEdxPublicSignal.instances = []
        with pytest.raises(KeyError):
            OpenEdxPublicSignal.get_signal_by_type('org.openedx.content_authoring.course.catalog_info.changed.v1')
        with pytest.raises(KeyError):
            OpenEdxPublicSignal.get_signal_by_type('org.openedx.learning.course.enrollment.created.v1')

        load_all_signals()
        assert isinstance(
            OpenEdxPublicSignal.get_signal_by_type('org.openedx.content_authoring.course.catalog_info.changed.v1'),
            OpenEdxPublicSignal
        )
        assert isinstance(
            OpenEdxPublicSignal.get_signal_by_type('org.openedx.learning.course.enrollment.created.v1'),
            OpenEdxPublicSignal
        )



"""
Utility methods and classes for testing various modules in event_bus.avro.
"""
import re
from datetime import datetime

import attr
from opaque_keys.edx.keys import CourseKey

from openedx_events.event_bus.avro.custom_serializers import BaseCustomTypeAvroSerializer
from openedx_events.event_bus.avro.deserializer import AvroSignalDeserializer
from openedx_events.event_bus.avro.serializer import AvroSignalSerializer
from openedx_events.event_bus.avro.types import PYTHON_TYPE_TO_AVRO_MAPPING
from openedx_events.tooling import OpenEdxPublicSignal


def create_simple_signal(data_dict, event_type="simple.signal"):
    """
    Create a basic OpenEdxPublicSignal with init_data = data_dict.

    Arguments:
        data_dict: Description of attributes passed to the signal
        event_type: A custom event type string. Defaults to 'simple.signal'
    """
    return OpenEdxPublicSignal(
        event_type=event_type,
        data=data_dict
    )


# Useful simple attr classes
@attr.s(auto_attribs=True)
class SimpleAttrs:
    """Class with all primitive type fields"""
    boolean_field: bool
    int_field: int
    float_field: float
    bytes_field: bytes
    string_field: str


@attr.s(auto_attribs=True)
class SubTestData0:
    """Subclass for testing nested attrs"""
    sub_name: str
    course_id: str


@attr.s(auto_attribs=True)
class SubTestData1:
    """Subclass for testing nested attrs"""
    sub_name: str
    course_id: str


@attr.s(auto_attribs=True)
class EventData:
    """More complex class for testing nested attrs"""
    sub_name: str
    course_id: str
    sub_test_0: SubTestData0
    sub_test_1: SubTestData1


@attr.s(frozen=True)
class SimpleAttrsWithDefaults:
    """Test attrs with nullable values"""
    boolean_field = attr.ib(type=bool, default=None)
    int_field = attr.ib(type=int, default=None)
    float_field = attr.ib(type=float, default=None)
    bytes_field = attr.ib(type=bytes, default=None)
    string_field = attr.ib(type=str, default=None)
    attrs_field = attr.ib(type=SimpleAttrs, default=None)


@attr.s(frozen=True)
class CustomAttrsWithDefaults:
    """Test attrs with nullable values"""
    coursekey_field = attr.ib(type=CourseKey, default=None)
    datetime_field = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class CustomAttrsWithoutDefaults:
    """Test attrs without nullable values"""
    coursekey_field = attr.ib(type=CourseKey)
    datetime_field = attr.ib(type=datetime)


@attr.s(frozen=True)
class NestedAttrsWithDefaults:
    """Test attrs with nullable values"""
    field_0 = attr.ib(type=SimpleAttrsWithDefaults)


class NonAttrs:
    """Test data class not decorated with @attr."""
    def __init__(self, val0, val1):
        self.val0 = val0
        self.val1 = val1

    def __eq__(self, other):
        # Treat all instances with the same values as equal for easier testing
        return self.val0 == other.val0 and self.val1 == other.val1


@attr.s(frozen=True)
class NestedNonAttrs:
    """Test attrs with nullable values"""
    field_0 = attr.ib(type=NonAttrs)


class NonAttrsAvroSerializer(BaseCustomTypeAvroSerializer):
    """Custom serializer for Non-Attrs class"""

    cls = NonAttrs
    field_type = PYTHON_TYPE_TO_AVRO_MAPPING[str]

    @staticmethod
    def serialize(obj) -> str:
        return f"{obj.val0}:{obj.val1}"

    @staticmethod
    def deserialize(data: str) -> object:
        bits = re.split(":", data)
        return NonAttrs(bits[0], bits[1])


class SpecialSerializer(AvroSignalSerializer):
    def custom_type_serializers(self):
        return [NonAttrsAvroSerializer]


class SpecialDeserializer(AvroSignalDeserializer):
    def custom_type_serializers(self):
        return [NonAttrsAvroSerializer]



"""
Test for producer configuration.
"""
from unittest.mock import Mock, patch

import ddt
import pytest
from django.apps import apps
from django.test import TestCase, override_settings

from openedx_events.content_authoring.data import XBlockData
from openedx_events.content_authoring.signals import XBLOCK_DELETED, XBLOCK_PUBLISHED
from openedx_events.exceptions import ProducerConfigurationError


@ddt.ddt
class ProducerConfiguratonTest(TestCase):
    """
    Tests to make sure EVENT_BUS_PRODUCER_CONFIG setting connects required signals to appropriate handlers.

    Attributes:
        xblock_info: dummy XBlockData.
    """
    def setUp(self) -> None:
        super().setUp()
        self.xblock_info = XBlockData(
            usage_key='block-v1:edx+DemoX+Demo_course+type@video+block@UaEBjyMjcLW65gaTXggB93WmvoxGAJa0JeHRrDThk',
            block_type='video',
        )

    @patch('openedx_events.apps.get_producer')
    def test_enabled_disabled_events(self, mock_producer):
        """
        Check whether XBLOCK_PUBLISHED is connected to the handler and the handler only produces enabled events.

        Args:
            mock_producer: mock get_producer to inspect the arguments.
        """
        mock_send = Mock()
        mock_producer.return_value = mock_send
        # XBLOCK_PUBLISHED has three configurations where 2 configurations have set enabled as True.
        XBLOCK_PUBLISHED.send_event(xblock_info=self.xblock_info)
        mock_send.send.assert_called()
        mock_send.send.call_count = 2

        # check that call_args_list only consists of enabled topics.
        call_args = mock_send.send.call_args_list[0][1]
        self.assertDictContainsSubset(
            {'topic': 'enabled_topic_a', 'event_key_field': 'xblock_info.usage_key'},
            call_args
        )
        call_args = mock_send.send.call_args_list[1][1]
        self.assertDictContainsSubset(
            {'topic': 'enabled_topic_b', 'event_key_field': 'xblock_info.usage_key'},
            call_args
        )

    @patch('openedx_events.apps.get_producer')
    @override_settings(EVENT_BUS_PRODUCER_CONFIG={})
    def test_events_not_in_config(self, mock_producer):
        """
        Check whether events not included in the configuration are not published as expected.

        Args:
            mock_producer: mock get_producer to inspect the arguments.
        """
        mock_send = Mock()
        mock_producer.return_value = mock_send
        XBLOCK_PUBLISHED.send_event(xblock_info=self.xblock_info)
        mock_producer.assert_not_called()
        mock_send.send.assert_not_called()

    def test_configuration_is_validated(self):
        """
        Check whether EVENT_BUS_PRODUCER_CONFIG setting is validated before connecting handlers.
        """
        with override_settings(EVENT_BUS_PRODUCER_CONFIG=[]):
            with pytest.raises(ProducerConfigurationError, match="should be a dictionary"):
                apps.get_app_config("openedx_events").ready()

        with override_settings(EVENT_BUS_PRODUCER_CONFIG={"invalid.event.type": {}}):
            with pytest.raises(ProducerConfigurationError, match="No OpenEdxPublicSignal of type"):
                apps.get_app_config("openedx_events").ready()

        with override_settings(EVENT_BUS_PRODUCER_CONFIG={"org.openedx.content_authoring.xblock.deleted.v1": ""}):
            with pytest.raises(ProducerConfigurationError, match="should be a dict"):
                apps.get_app_config("openedx_events").ready()

        with override_settings(EVENT_BUS_PRODUCER_CONFIG={"org.openedx.content_authoring.xblock.deleted.v1":
                                                          {"topic": ""}}):
            with pytest.raises(ProducerConfigurationError, match="One of the configuration objects is not a"
                                                                 " dictionary"):
                apps.get_app_config("openedx_events").ready()

        with override_settings(
            EVENT_BUS_PRODUCER_CONFIG={
                "org.openedx.content_authoring.xblock.deleted.v1": {"topic": {"enabled": True}}
            }
        ):
            with pytest.raises(ProducerConfigurationError, match="missing 'event_key_field' key."):
                apps.get_app_config("openedx_events").ready()

        with override_settings(
            EVENT_BUS_PRODUCER_CONFIG={
                "org.openedx.content_authoring.xblock.deleted.v1":
                {
                    "some": {"enabled": 1, "event_key_field": "some"}
                }
            }
        ):
            with pytest.raises(
                ProducerConfigurationError,
                match="Expected type: <class 'bool'> for 'enabled', found: <class 'int'>"
            ):
                apps.get_app_config("openedx_events").ready()

    @patch('openedx_events.apps.get_producer')
    def test_event_data_key_in_handler(self, mock_producer):
        """
        Check whether event_data is constructed properly in handlers.
        """
        mock_send = Mock()
        mock_producer.return_value = mock_send
        XBLOCK_DELETED.send_event(xblock_info=self.xblock_info)
        mock_send.send.assert_called_once()

        call_args = mock_send.send.call_args_list[0][1]
        self.assertIn("xblock_info", call_args["event_data"])



"""
Utils used by Open edX event tests.
"""

from openedx_events.tooling import OpenEdxPublicSignal


class FreezeSignalCacheMixin:
    """
    A mixin to be used by TestCases to avoid new signals persisting in the OpenEdxPublicSignal cache of instances.
    """

    @classmethod
    def setUpClass(cls):
        """
        Save current signal instances.
        """
        super().setUpClass()
        cls.pre_run_instances = list(OpenEdxPublicSignal.instances)
        cls.pre_run_mapping = dict(OpenEdxPublicSignal._mapping)  # pylint: disable=protected-access

    @classmethod
    def tearDownClass(cls):
        """
        Restore instance cache to pre-test state.
        """
        super().tearDownClass()
        OpenEdxPublicSignal.instances = cls.pre_run_instances
        OpenEdxPublicSignal._mapping = cls.pre_run_mapping  # pylint: disable=protected-access


class EventsIsolationMixin:
    """
    A mixin to be used by TestCases that want to isolate their use of Open edX Events.
    """

    @classmethod
    def disable_all_events(cls):
        """
        Disable all events Open edX Events from all subdomains.
        """
        for event in OpenEdxPublicSignal.all_events():
            event.disable()

    @classmethod
    def enable_all_events(cls):
        """
        Enable all events Open edX Events from all subdomains.
        """
        for event in OpenEdxPublicSignal.all_events():
            event.enable()

    @classmethod
    def enable_events_by_type(cls, *event_types):
        """
        Enable specific Open edX Events given their type.

        Arguments:
            event_types (list of `str`): types of events to enable.
        """
        for event_type in event_types:
            try:
                event = OpenEdxPublicSignal.get_signal_by_type(event_type)
            except KeyError:
                all_event_types = sorted(s.event_type for s in OpenEdxPublicSignal.all_events())
                err_msg = (
                    "You tried to enable event '{}', but I don't recognize that "
                    "signal type. Did you mean one of these?: {}"
                )
                raise ValueError(err_msg.format(event_type, all_event_types))  # pylint: disable=raise-missing-from
            event.enable()

    @classmethod
    def allow_send_events_failure(cls, *event_types):
        """
        Allow that send_event method fails for the specified event.

        This method determines which `send` method to use, send or send_robust.
        The first, raises receivers exceptions while the latter catches them.

        Arguments:
            event_types (list of `str`): types of events to enable.
        """
        for event_type in event_types:
            try:
                event = OpenEdxPublicSignal.get_signal_by_type(event_type)
            except KeyError:
                all_event_types = sorted(s.event_type for s in OpenEdxPublicSignal.all_events())
                err_msg = (
                    "You tried to enable event '{}', but I don't recognize that "
                    "signal type. Did you mean one of these?: {}"
                )
                raise ValueError(err_msg.format(event_type, all_event_types))  # pylint: disable=raise-missing-from
            event.allow_send_event_failure()


class OpenEdxEventsTestMixin(EventsIsolationMixin):
    """
    A mixin to be used by TestCases that want to isolate their use of Open edX Events.

    Example usage:

        class MyTestCase(TestCase, OpenEdxEventsTestCase):

            ENABLED_OPENEDX_EVENTS = ['org.openedx.learning.student.registration.completed.v1']

    This class assumes that's it's being used in conjunction TestCase or TestCase subclasses.
    """

    ENABLED_OPENEDX_EVENTS = []

    @classmethod
    def setUpClass(cls):
        """
        Start events isolation for class.
        """
        super().setUpClass()
        cls().start_events_isolation()

    @classmethod
    def start_events_isolation(cls):
        """
        Start Open edX Events isolation and then enable events by type.
        """
        cls().disable_all_events()
        cls().enable_events_by_type(*cls.ENABLED_OPENEDX_EVENTS)
        cls().allow_send_events_failure(*cls.ENABLED_OPENEDX_EVENTS)



"""
Tests for event bus implementation loader.
"""

import copy
import warnings
from contextlib import contextmanager
from unittest import TestCase

from django.test import override_settings

from openedx_events.data import EventsMetadata
from openedx_events.event_bus import _try_load, get_producer, make_single_consumer, merge_producer_configs
from openedx_events.learning.signals import SESSION_LOGIN_COMPLETED


@contextmanager
def assert_warnings(warning_messages: list):
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')
        yield
    assert [str(w.message) for w in caught_warnings] == warning_messages


class TestLoader(TestCase):

    # No, the "constructors" here don't make much sense, but I didn't
    # want to create a bunch of test classes/factory functions, so I'm
    # using built-in functions instead.

    def test_unconfigured(self):
        with assert_warnings(["Event Bus setting DOES_NOT_EXIST is missing; component will be inactive"]):
            loaded = _try_load(
                setting_name="DOES_NOT_EXIST", args=(1), kwargs={'2': 3},
                expected_class=dict, default={'def': 'ault'},
            )
        assert loaded == {'def': 'ault'}

    @override_settings(EB_LOAD_PATH='builtins.dict')
    def test_success(self):
        with assert_warnings([]):
            loaded = _try_load(
                setting_name="EB_LOAD_PATH", args=(), kwargs={'2': 3},
                expected_class=dict, default={'def': 'ault'},
            )
        assert loaded == {'2': 3}

    @override_settings(EB_LOAD_PATH='builtins.list')
    def test_wrong_type(self):
        with assert_warnings([
                "builtins.list from EB_LOAD_PATH returned unexpected type <class 'list'>; "
                "component will be inactive"
        ]):
            loaded = _try_load(
                setting_name="EB_LOAD_PATH", args=([1, 2, 3],), kwargs={},
                expected_class=dict, default={'def': 'ault'},
            )
        assert loaded == {'def': 'ault'}

    @override_settings(EB_LOAD_PATH='no_module_here.foo.nope')
    def test_missing_module(self):
        with assert_warnings([
                "Failed to load <class 'dict'> from setting EB_LOAD_PATH: "
                "ModuleNotFoundError(\"No module named 'no_module_here'\"); "
                "component will be inactive"
        ]):
            loaded = _try_load(
                setting_name="EB_LOAD_PATH", args=(1), kwargs={'2': 3},
                expected_class=dict, default={'def': 'ault'},
            )
        assert loaded == {'def': 'ault'}

    @override_settings(EB_LOAD_PATH='builtins.does_not_exist')
    def test_missing_attribute(self):
        with assert_warnings([
                "Failed to load <class 'dict'> from setting EB_LOAD_PATH: "
                "ImportError('Module \"builtins\" does not define a \"does_not_exist\" attribute/class'); "
                "component will be inactive"
        ]):
            loaded = _try_load(
                setting_name="EB_LOAD_PATH", args=(1), kwargs={'2': 3},
                expected_class=dict, default={'def': 'ault'},
            )
        assert loaded == {'def': 'ault'}

    @override_settings(EB_LOAD_PATH='builtins.dict')
    def test_bad_args_for_callable(self):
        with assert_warnings([
                "Failed to load <class 'dict'> from setting EB_LOAD_PATH: "
                "TypeError('type object argument after * must be an iterable, not int'); "
                "component will be inactive"
        ]):
            loaded = _try_load(
                setting_name="EB_LOAD_PATH", args=(1), kwargs={'2': 3},
                expected_class=dict, default={'def': 'ault'},
            )
        assert loaded == {'def': 'ault'}


class TestProducer(TestCase):

    @override_settings(EVENT_BUS_PRODUCER=None)
    def test_default_does_nothing(self):
        """
        Test that the default is of the right class but does nothing.
        """
        producer = get_producer()

        with assert_warnings([]):
            # Nothing thrown, no warnings.
            assert producer.send(
                signal=SESSION_LOGIN_COMPLETED, topic='user-logins',
                event_key_field='user.id', event_data={},
                event_metadata=EventsMetadata(event_type='eh')
            ) is None


class TestConsumer(TestCase):

    @override_settings(EVENT_BUS_CONSUMER=None)
    def test_default_does_nothing(self):
        """
        Test that the default is of the right class but does nothing.
        """
        consumer = make_single_consumer(topic="test", group_id="test", signal=SESSION_LOGIN_COMPLETED)

        with assert_warnings([]):
            # Nothing thrown, no warnings.
            assert consumer.consume_indefinitely() is None


class TestSettings(TestCase):

    def setUp(self) -> None:
        super().setUp()
        self.base_config = {
            'event_type_0': {
                'topic_a': {'event_key_field': 'field', 'enabled': True},
                'topic_b': {'event_key_field': 'field', 'enabled': True}
            },
            'event_type_1': {
                'topic_c': {'event_key_field': 'field', 'enabled': True},
            }
        }

    def test_merge_configs(self):
        # for ensuring we didn't change the original dict
        base_copy = copy.deepcopy(self.base_config)
        overrides = {
            'event_type_0': {
                # disable an existing event/topic pairing and change the key field
                'topic_a': {'event_key_field': 'new_field', 'enabled': False},
                # add a new topic to an existing event_type
                'topic_d': {'event_key_field': 'field', 'enabled': True},
            },
            # add a new event_type
            'event_type_2': {
                'topic_e': {'event_key_field': 'field', 'enabled': True},
            }
        }
        overrides_copy = copy.deepcopy(overrides)
        result = merge_producer_configs(self.base_config, overrides)
        self.assertDictEqual(result, {
            'event_type_0': {
                'topic_a': {'event_key_field': 'new_field', 'enabled': False},
                'topic_b': {'event_key_field': 'field', 'enabled': True},
                'topic_d': {'event_key_field': 'field', 'enabled': True},
            },
            'event_type_1': {
                'topic_c': {'event_key_field': 'field', 'enabled': True},
            },
            'event_type_2': {
                'topic_e': {'event_key_field': 'field', 'enabled': True},
            }
        })
        self.assertDictEqual(self.base_config, base_copy)
        self.assertDictEqual(overrides, overrides_copy)

    def test_merge_configs_with_empty(self):
        overrides = {}
        result = merge_producer_configs(self.base_config, overrides)
        self.assertDictEqual(result, self.base_config)

    def test_merge_configs_with_partial(self):
        overrides = {
            'event_type_0': {
                # no override for 'event_key_field'
                'topic_a': {'enabled': False},
                # no override for 'enabled'
                'topic_b': {'event_key_field': 'new_field'}
            }
        }
        result = merge_producer_configs(self.base_config, overrides)
        self.assertDictEqual(result, {
            'event_type_0': {
                'topic_a': {'event_key_field': 'field', 'enabled': False},
                'topic_b': {'event_key_field': 'new_field', 'enabled': True}
            },
            'event_type_1': {
                'topic_c': {'event_key_field': 'field', 'enabled': True},
            }
        })



""" Tests for openedx_events.data module."""
from datetime import datetime, timezone
from uuid import UUID

import ddt
from django.test import TestCase
from django.test.utils import override_settings

from openedx_events.data import EventsMetadata


@ddt.ddt
class TestEventsMetadata(TestCase):
    """
    Tests for the EventsMetadata class.
    """
    def setUp(self) -> None:
        self.now = datetime.now(timezone.utc)
        self.metadata = EventsMetadata(
            event_type='test_type',
            time=self.now,
            source='test_source',
            sourcehost='test_source_host',
            sourcelib=(1, 2, 3),
            id=UUID('c45efb10-3556-11ee-9f19-7e694b1e500b')
        )

    def test_events_metadata_to_and_from_json(self):
        as_json = self.metadata.to_json()
        from_json = EventsMetadata.from_json(as_json)
        self.assertEqual(self.metadata, from_json)

    @ddt.data(
        ('settings_variant', None, 'openedx/settings_variant/web'),
        (None, 'my_service', 'openedx/my_service/web'),
        (None, None, 'openedx/SERVICE_NAME_UNSET/web'),
        ('settings_variant', 'my_service', 'openedx/my_service/web')
    )
    @ddt.unpack
    def test_events_metadata_source(self, settings_variant, event_bus_service_name, expected_source):
        with override_settings(
                SERVICE_VARIANT=settings_variant,
                EVENTS_SERVICE_NAME=event_bus_service_name,
        ):
            metadata = EventsMetadata(
                event_type='test_type'
            )
            self.assertEqual(metadata.source, expected_source)



"""
Data attributes for events within the architecture subdomain ``analytics``.

These attributes follow the form of attr objects specified in OEP-49 data
pattern.
"""

from datetime import datetime

import attr


@attr.s(frozen=True)
class TrackingLogData:
    """
    Data describing tracking events.

    Arguments:
        name (str): event name
        timestamp (datetime): timestamp of the event
        data (str): json string representation of a dictionary with extra data (optional),
                    e.g. {"course_id": "course-v1:edX+DemoX+Demo_Course"}
        context (dict): json string representation of a dictionary of context data
            defined in https://edx.readthedocs.io/projects/devdata/en/latest/internal_data_formats/tracking_logs/
    """

    name = attr.ib(type=str)
    timestamp = attr.ib(type=datetime)
    data = attr.ib(type=str)
    context = attr.ib(type=str)



"""Tests for avro.deserializer"""
import json
from datetime import datetime
from typing import List
from unittest import TestCase

from opaque_keys.edx.keys import CourseKey, UsageKey
from opaque_keys.edx.locator import LibraryLocatorV2, LibraryUsageLocatorV2

from openedx_events.event_bus.avro.deserializer import AvroSignalDeserializer, deserialize_bytes_to_event_data
from openedx_events.event_bus.avro.tests.test_utilities import (
    EventData,
    NestedAttrsWithDefaults,
    NestedNonAttrs,
    NonAttrs,
    SimpleAttrs,
    SimpleAttrsWithDefaults,
    SpecialDeserializer,
    SubTestData0,
    SubTestData1,
    create_simple_signal,
)
from openedx_events.tests.utils import FreezeSignalCacheMixin


class TestAvroSignalDeserializerCache(TestCase, FreezeSignalCacheMixin):
    """Test AvroSignalDeserializer"""

    def setUp(self) -> None:
        super().setUp()
        self.maxDiff = None

    def test_schema_string(self):
        """
        Test JSON round-trip; schema creation is tested more fully in test_schema.py.
        """
        SIGNAL = create_simple_signal({
            "data": SimpleAttrs
        })
        actual_schema = json.loads(AvroSignalDeserializer(SIGNAL).schema_string())
        expected_schema = {
            'name': 'CloudEvent',
            'type': 'record',
            'doc': 'Avro Event Format for CloudEvents created with openedx_events/schema',
            'namespace': 'simple.signal',
            'fields': [
                {
                    'name': 'data',
                    'type': {
                        'name': 'SimpleAttrs',
                        'type': 'record',
                        'fields': [
                            {'name': 'boolean_field', 'type': 'boolean'},
                            {'name': 'int_field', 'type': 'long'},
                            {'name': 'float_field', 'type': 'double'},
                            {'name': 'bytes_field', 'type': 'bytes'},
                            {'name': 'string_field', 'type': 'string'},
                        ]
                    }
                }
            ]
        }
        assert actual_schema == expected_schema

    def test_convert_dict_to_event_data(self):
        SIGNAL = create_simple_signal({"test_data": EventData})
        initial_dict = {
            "test_data": {
                "course_id": "bar.course",
                "sub_name": "foo",
                "sub_test_0": {"course_id": "a.nother.course", "sub_name": "a.sub.name"},
                "sub_test_1": {
                    "course_id": "b.uber.another.course",
                    "sub_name": "b.uber.sub.name",
                },
            }
        }
        deserializer = AvroSignalDeserializer(SIGNAL)
        event_data = deserializer.from_dict(initial_dict)
        expected_event_data = EventData("foo", "bar.course",
                                        SubTestData0("a.sub.name", "a.nother.course"),
                                        SubTestData1("b.uber.sub.name", "b.uber.another.course")
                                        )
        test_data = event_data["test_data"]
        self.assertIsInstance(test_data, EventData)
        self.assertEqual(test_data, expected_event_data)

    def test_default_datetime_deserialization(self):
        SIGNAL = create_simple_signal({"birthday": datetime})
        deserializer = AvroSignalDeserializer(SIGNAL)
        birthday = datetime(year=1989, month=9, day=6)
        as_dict = {"birthday": birthday.isoformat()}
        event_data = deserializer.from_dict(as_dict)
        birthday_deserialized = event_data["birthday"]
        self.assertIsInstance(birthday_deserialized, datetime)
        self.assertEqual(birthday_deserialized, birthday)

    def test_default_coursekey_deserialization(self):
        SIGNAL = create_simple_signal({"course": CourseKey})
        deserializer = AvroSignalDeserializer(SIGNAL)
        course_key = CourseKey.from_string("course-v1:edX+DemoX.1+2014")
        as_dict = {"course": str(course_key)}
        event_data = deserializer.from_dict(as_dict)
        course_deserialized = event_data["course"]
        self.assertIsInstance(course_deserialized, CourseKey)
        self.assertEqual(course_deserialized, course_key)

    def test_default_usagekey_deserialization(self):
        """
        Test deserialization of UsageKey
        """
        SIGNAL = create_simple_signal({"usage_key": UsageKey})
        deserializer = AvroSignalDeserializer(SIGNAL)
        usage_key = UsageKey.from_string(
            "block-v1:edx+DemoX+Demo_course+type@video+block@UaEBjyMjcLW65gaTXggB93WmvoxGAJa0JeHRrDThk",
        )
        as_dict = {"usage_key": str(usage_key)}
        event_data = deserializer.from_dict(as_dict)
        usage_key_deserialized = event_data["usage_key"]
        self.assertIsInstance(usage_key_deserialized, UsageKey)
        self.assertEqual(usage_key_deserialized, usage_key)

    def test_default_librarylocatorv2_deserialization(self):
        """
        Test deserialization of LibraryLocatorV2
        """
        SIGNAL = create_simple_signal({"library_key": LibraryLocatorV2})
        deserializer = AvroSignalDeserializer(SIGNAL)
        library_key = LibraryLocatorV2.from_string("lib:MITx:reallyhardproblems")
        as_dict = {"library_key": str(library_key)}
        event_data = deserializer.from_dict(as_dict)
        library_key_deserialized = event_data["library_key"]
        self.assertIsInstance(library_key_deserialized, LibraryLocatorV2)
        self.assertEqual(library_key_deserialized, library_key)

    def test_default_libraryusagelocatorv2_deserialization(self):
        """
        Test deserialization of LibraryUsageLocatorV2
        """
        SIGNAL = create_simple_signal({"usage_key": LibraryUsageLocatorV2})
        deserializer = AvroSignalDeserializer(SIGNAL)
        usage_key = LibraryUsageLocatorV2.from_string("lb:MITx:reallyhardproblems:problem:problem1")
        as_dict = {"usage_key": str(usage_key)}
        event_data = deserializer.from_dict(as_dict)
        usage_key_deserialized = event_data["usage_key"]
        self.assertIsInstance(usage_key_deserialized, LibraryUsageLocatorV2)
        self.assertEqual(usage_key_deserialized, usage_key)

    def test_deserialization_with_custom_serializer(self):
        SIGNAL = create_simple_signal({"test_data": NonAttrs})
        deserializer = SpecialDeserializer(SIGNAL)
        as_dict = {
            "test_data": "a.val:a.nother.val"
        }
        event_data = deserializer.from_dict(as_dict)
        non_attrs_deserialized = event_data["test_data"]
        self.assertIsInstance(non_attrs_deserialized, NonAttrs)
        self.assertEqual(non_attrs_deserialized, NonAttrs("a.val", "a.nother.val"))

    def test_deserialization_with_custom_serializer_on_nested_fields(self):
        SIGNAL = create_simple_signal({"test_data": NestedNonAttrs})
        as_dict = {"test_data": {
            "field_0": "a.val:a.nother.val"
        }}
        deserializer = SpecialDeserializer(SIGNAL)
        event_data = deserializer.from_dict(as_dict)
        deserialized_nested = event_data["test_data"]
        self.assertIsInstance(deserialized_nested, NestedNonAttrs)
        inner_deserialized_non_attrs = deserialized_nested.field_0
        self.assertIsInstance(inner_deserialized_non_attrs, NonAttrs)
        self.assertEqual(inner_deserialized_non_attrs, NonAttrs("a.val", "a.nother.val"))

    def test_deserialization_fails_if_missing_fields(self):
        # missing "sub_name" field
        initial_dict = {
            "test_data": {
                "course_id": "bar.course",
            }
        }

        SIGNAL = create_simple_signal({"test_data": SubTestData0})
        deserializer = AvroSignalDeserializer(SIGNAL)
        with self.assertRaises(TypeError):
            _ = deserializer.from_dict(initial_dict)

    def test_deserialization_of_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": SimpleAttrsWithDefaults
        })
        deserializer = AvroSignalDeserializer(SIGNAL)

        as_dict = {"data": {}}
        data_dict = deserializer.from_dict(as_dict)
        self.assertIsInstance(data_dict["data"], SimpleAttrsWithDefaults)
        self.assertEqual(data_dict["data"], SimpleAttrsWithDefaults())

    def test_deserialization_of_nested_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": NestedAttrsWithDefaults
        })
        deserializer = AvroSignalDeserializer(SIGNAL)
        as_dict = {"data": {"field_0": {}}}
        data_dict = deserializer.from_dict(as_dict)
        nested_field = data_dict["data"].field_0
        self.assertIsInstance(nested_field, SimpleAttrsWithDefaults)
        self.assertEqual(nested_field, SimpleAttrsWithDefaults())

    def test_deserialization_of_list_with_annotation(self):
        """
        Check that deserialization works as expected when list data is annotated.
        """
        LIST_SIGNAL = create_simple_signal({"list_input": List[int]})
        initial_dict = {"list_input": [1, 3]}
        deserializer = AvroSignalDeserializer(LIST_SIGNAL)
        event_data = deserializer.from_dict(initial_dict)
        expected_event_data = [1, 3]
        test_data = event_data["list_input"]
        self.assertIsInstance(test_data, list)
        self.assertEqual(test_data, expected_event_data)

    def test_deserialization_of_list_without_annotation(self):
        """
        Check that deserialization raises error when list data is not annotated.
        """
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        SIGNAL = create_simple_signal({"list_input": List[int]})
        LIST_SIGNAL = create_simple_signal({"list_input": List})
        initial_dict = {"list_input": [1, 3]}
        deserializer = AvroSignalDeserializer(SIGNAL)
        # Update signal with incomplete type info
        deserializer.signal = LIST_SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_nested_list_fails(self):
        """
        Check that deserialization raises error when nested list data is passed.
        """
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        SIGNAL = create_simple_signal({"list_input": List[int]})
        LIST_SIGNAL = create_simple_signal({"list_input": List[List[int]]})
        initial_dict = {"list_input": [[1, 3], [4, 5]]}
        deserializer = AvroSignalDeserializer(SIGNAL)
        # Update signal with incomplete type info
        deserializer.signal = LIST_SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialization_of_nested_list_with_complex_types_fails(self):
        SIGNAL = create_simple_signal({"list_input": List[list]})
        with self.assertRaises(TypeError):
            AvroSignalDeserializer(SIGNAL)
        initial_dict = {"list_input": [[1, 3], [4, 5]]}
        # create dummy signal to bypass schema check while initializing deserializer
        # This allows us to test whether correct exceptions are raised while deserializing data
        DUMMY_SIGNAL = create_simple_signal({"list_input": List[int]})
        deserializer = AvroSignalDeserializer(DUMMY_SIGNAL)
        # Update signal with incorrect type info
        deserializer.signal = SIGNAL
        with self.assertRaises(TypeError):
            deserializer.from_dict(initial_dict)

    def test_deserialize_bytes_to_event_data(self):
        """
        Test deserialize_bytes_to_event_data utility function.
        """
        SIGNAL = create_simple_signal({"test_data": EventData})
        bytes_data = b'\x06foo\x14bar.course\x14a.sub.name\x1ea.nother.course\x1eb.uber.sub.name*b.uber.another.course'
        expected = {"test_data": EventData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )}
        deserialized = deserialize_bytes_to_event_data(bytes_data, SIGNAL)
        self.assertIsInstance(deserialized["test_data"], EventData)
        self.assertEqual(deserialized, expected)



"""
Package where Open edX Events and the necessary tooling are implemented.

These definitions are part of the Hooks Extension Framework, see OEP-50 for
more information about the project.
"""

__version__ = "9.5.0"



"""
Tooling necessary to use Open edX events.
"""
import pkgutil
import warnings
from importlib import import_module
from logging import getLogger

from django.conf import settings
from django.db import connection
from django.dispatch import Signal
from edx_django_utils.cache import RequestCache

from openedx_events.data import EventsMetadata
from openedx_events.exceptions import SenderValidationError
from openedx_events.utils import format_responses

log = getLogger(__name__)

# If a signal is explicitly not for use with the event bus, add it to this list
#  and document why in the event's annotations
KNOWN_UNSERIALIZABLE_SIGNALS = [
    "org.openedx.learning.discussions.configuration.changed.v1",
    "org.openedx.content_authoring.course.certificate_config.changed.v1",
    "org.openedx.content_authoring.course.certificate_config.deleted.v1",
    "org.openedx.learning.user.notification.requested.v1",
    "org.openedx.learning.thread.created.v1",
    "org.openedx.learning.response.created.v1",
    "org.openedx.learning.comment.created.v1",
    "org.openedx.learning.course.notification.requested.v1",
]


class OpenEdxPublicSignal(Signal):
    """
    Standardized Django Signals used to create Open edX events.
    """

    _mapping = {}
    instances = []

    def __init__(self, event_type, data, minor_version=0):
        """
        Init method for OpenEdxPublicSignal definition class.

        Arguments:
            event_type (str): name of the event.
            data (dict): attributes passed to the event.
            minor_version (int): version of the event type.
        """
        self.init_data = data
        self.event_type = event_type
        self.minor_version = minor_version
        self._allow_events = True
        self._allow_send_event_failure = False
        self.__class__.instances.append(self)
        self.__class__._mapping[self.event_type] = self
        super().__init__()

    def __repr__(self):
        """
        Represent OpenEdxPublicSignal as a string.
        """
        return "<OpenEdxPublicSignal: {event_type}>".format(event_type=self.event_type)

    @classmethod
    def all_events(cls):
        """
        Get all current events.
        """
        return cls.instances

    @classmethod
    def get_signal_by_type(cls, event_type):
        """
        Get event identified by type.

        Arguments:
            event_type (str): name of the event.

        Exceptions raised:
            Raises KeyError if not found.
        """
        return cls._mapping[event_type]

    def generate_signal_metadata(self, time=None):
        """
        Generate signal metadata when an event is sent.

        These fields are generated on the fly and are a subset of the Event
        Message defined in the OEP-41.

        Arguments:
            time (datetime): (Optional) Timestamp when the event was sent with
                UTC timezone. Defaults to current time in UTC. See OEP-41 for
                details.

        Example usage:
            >>> metadata = \
                STUDENT_REGISTRATION_COMPLETED.generate_signal_metadata()
                attr.asdict(metadata)
            {
                'event_type': '...learning.student.registration.completed.v1',
                'minorversion': 0,
                'time': '2021-06-09T14:12:45.320819Z',
                'source': 'openedx/lms/web',
                'sourcehost': 'edx.devstack.lms',
                'specversion': '1.0',
                'sourcelib: (0,1,0,),
            }
        """
        return EventsMetadata(
            event_type=self.event_type,
            minorversion=self.minor_version,
            time=time,
        )

    def _send_event_with_metadata(self, metadata, send_robust=True, **kwargs):
        """
        Send events to all connected receivers with the provided metadata.

        This method is for internal use only.

        Arguments:
            metadata (EventsMetadata): The metadata to be sent with the signal.
            send_robust (bool): Defaults to True. See Django signal docs.

        See ``send_event`` docstring for more details on its usage and behavior.
        """

        def validate_sender():
            """
            Run validations over the send arguments.

            The validation checks whether the send arguments match the
            arguments used when instantiating the event. If they don't a
            validation error is raised.
            """
            if len(kwargs) != len(self.init_data):
                raise SenderValidationError(
                    event_type=self.event_type,
                    message="There's a mismatch between initialization data and send_event arguments",
                )

            for key, value in self.init_data.items():
                argument = kwargs.get(key)
                if not argument:
                    raise SenderValidationError(
                        event_type=self.event_type,
                        message="Missing required argument '{key}'".format(key=key),
                    )
                if not isinstance(argument, value):
                    raise SenderValidationError(
                        event_type=self.event_type,
                        message="The argument '{key}' is not instance of the Class Attribute '{attr}'".format(
                            key=key, attr=value.__class__.__name__
                        ),
                    )

        if not self._allow_events:
            return []

        validate_sender()

        kwargs["metadata"] = metadata

        if self._allow_send_event_failure or settings.DEBUG or not send_robust:
            return super().send(sender=None, **kwargs)

        responses = super().send_robust(sender=None, **kwargs)
        log.info(
            f"Responses of the Open edX Event <{self.event_type}>: \n{format_responses(responses, depth=2)}",
        )

        return responses

    def send_event(self, send_robust=True, time=None, **kwargs):
        """
        Send events to all connected receivers.

        Arguments:
            send_robust (bool): Defaults to True. See Django signal docs.
            time (datetime): (Optional - see note) Timestamp when the event was
                sent with UTC timezone. For events requiring a DB create or
                update, use the timestamp from the DB record. Defaults to
                current time in UTC. This argument is optional for backward
                compatability, but ideally would be explicitly set. See OEP-41
                for details.
            kwargs: Data to be sent to the signal's receivers.

        Used to send events just like Django signals are sent. In addition,
        some validations are executed on the arguments, and then generates relevant
        metadata that can be used for logging or debugging purposes. Besides this behavior,
        send_event behaves just like the send method.

        If the event is disabled (i.e _allow_events is False), then this method
        won't have any effect. Meaning, the Django Signal won't be sent.

        Example usage:
            >>> STUDENT_REGISTRATION_COMPLETED.send_event(
                user=user_data, registration=registration_data,
            )
            [(<function callback at 0x7f2ce638ef70>, 'callback response')]

        Returns:
            list: response of each receiver following the format
            [(receiver, response), ... ]. Empty list if the event is disabled.

        Exceptions raised:
            SenderValidationError: raised when there's a mismatch between
            arguments passed to this method and arguments used to initialize
            the event.
        """
        metadata = self.generate_signal_metadata(time=time)
        return self._send_event_with_metadata(metadata=metadata, send_robust=send_robust, **kwargs)

    def send_event_with_custom_metadata(
            self, metadata, /, *, send_robust=True, **kwargs
    ):
        """
        Send events to all connected receivers using the provided metadata.

        This method works exactly like ``send_event``, except it uses the given
            event metadata rather than generating it. This is used by the
            event bus consumer, where we want to recreate the metadata used
            in the producer when resending the same signal on the consuming
            side.

        Arguments:
            metadata (EventsMetadata): The metadata to be sent with the signal.
            send_robust (bool): Defaults to True. See Django signal docs.
            kwargs: Data to be sent to the signal's receivers.

        See ``send_event`` docstring for more details.

        """
        return self._send_event_with_metadata(metadata=metadata, send_robust=send_robust, **kwargs)

    def send(self, sender, **kwargs):  # pylint: disable=unused-argument
        """
        Override method used to recommend the sender to adopt our custom send.
        """
        warnings.warn("Please, use 'send_event' when triggering an Open edX event.")

    def send_robust(self, sender, **kwargs):  # pylint: disable=unused-argument
        """
        Override method used to recommend the sender to adopt our custom send.
        """
        warnings.warn(
            "Please, use 'send_event' with send_robust equals to True when triggering an Open edX event."
        )

    def enable(self):
        """
        Enable all events. Meaning, send_event will send a Django signal.
        """
        self._allow_events = True

    def disable(self):
        """
        Disable all events. Meaning, send_event will have no effect.
        """
        self._allow_events = False

    def allow_send_event_failure(self):
        """
        Allow Django signal to fail. Meaning, uses send_robust instead of send.

        More information on send_robust in the Django official documentation.
        """
        self._allow_send_event_failure = True


def _process_all_signals_modules(func):
    """
    Walk the package tree and apply func on all signals.py files.

    Arguments:
        func: A method that takes a module name as its parameter
    """
    root = import_module('openedx_events')
    for m in pkgutil.walk_packages(root.__path__, root.__name__ + '.'):
        module_name = m.name
        if 'tests' in module_name.split('.') or '.test_' in module_name:
            continue
        if module_name.endswith('.signals'):
            func(module_name)


def load_all_signals():
    """
    Ensure OpenEdxPublicSignal.all_events() cache is fully populated.

    Loads all non-test signals.py modules.
    """
    _process_all_signals_modules(import_module)


def _reconnect_to_db_if_needed():  # pragma: no cover
    """
    Reconnects the db connection if needed.

    This is important because Django only does connection validity/age checks as part of
    its request/response cycle, which isn't in effect for the consume-loop. If we don't
    force these checks, a broken connection will remain broken indefinitely. For most
    consumers, this will cause event processing to fail.
    """
    has_connection = bool(connection.connection)
    requires_reconnect = has_connection and not connection.is_usable()
    if requires_reconnect:
        connection.connect()


def _clear_request_cache():  # pragma: no cover
    """
    Clear the RequestCache so that each event consumption starts fresh.

    Signal handlers may be written with the assumption that they are called in the context
    of a web request, so we clear the request cache just in case.
    """
    RequestCache.clear_all_namespaces()


def prepare_for_new_work_cycle():  # pragma: no cover
    """
    Ensure that the application state is appropriate for performing a new unit of work.

    This mimics some setup/teardown that is normally performed by Django in its
    request/response based architecture and that is needed for ensuring a clean and
    usable state in this worker-based application.

    See https://github.com/openedx/openedx-events/issues/236 for details.
    """
    # Ensure that the database connection is active and usable.
    _reconnect_to_db_if_needed()

    # Clear the request cache, in case anything in the signal handlers rely on it.
    _clear_request_cache()



"""
Code to convert attr classes to Avro specification.

TODO: Handle optional parameters and allow for schema evolution. https://github.com/edx/edx-arch-experiments/issues/53
"""


from typing import get_args, get_origin

from .custom_serializers import DEFAULT_CUSTOM_SERIALIZERS
from .types import PYTHON_TYPE_TO_AVRO_MAPPING, SIMPLE_PYTHON_TYPE_TO_AVRO_MAPPING

DEFAULT_FIELD_TYPES = {serializer.cls: serializer.field_type for serializer in DEFAULT_CUSTOM_SERIALIZERS}


def schema_from_signal(signal, custom_type_to_avro_type=None):
    """
    Create an Avro schema for events sent by an instance of OpenEdxPublicSignal.

    Arguments:
        signal: An instance of OpenEdxPublicSignal
        custom_type_to_avro_type: A map of Python class to Avro type
    """
    field_types = custom_type_to_avro_type or {}
    all_custom_field_types = {**DEFAULT_FIELD_TYPES, **field_types}
    previously_seen_types = set()

    base_schema = {
        "name": "CloudEvent",
        "type": "record",
        "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",
        "fields": [],
        "namespace": signal.event_type,
    }

    for data_key, data_type in signal.init_data.items():
        base_schema["fields"].append(_create_avro_field_definition(data_key, data_type,
                                                                   previously_seen_types,
                                                                   custom_type_to_avro_type=all_custom_field_types))
    return base_schema


def _create_avro_field_definition(data_key, data_type, previously_seen_types,
                                  custom_type_to_avro_type=None, default_is_none=False):
    """
    Create an Avro schema field definition from an OpenEdxPublicSignal data definition.

    Arguments:
        data_key: Field name, eg `lms_user_id`, `course_id`, `enrollment_status`
        data_type: Python data type, eg `str`, `CourseKey`, `CourseEnrollmentData`
        previously_seen_types: list of previously-encountered data types
        custom_type_to_avro_type: map of custom data types to a pre-determined avro field type
        default_is_none: boolean indicating whether this field has 'None' as a default
    """
    field = {"name": data_key}
    all_field_type_overrides = custom_type_to_avro_type or {}
    # get generic type of data_type
    # if data_type == List[int], data_type_origin = list
    data_type_origin = get_origin(data_type)

    # Case 1: data_type has a predetermined avro field representation
    if field_type := all_field_type_overrides.get(data_type, None):
        field["type"] = field_type
    # Case 2: data_type is a simple type that can be converted directly to an Avro type
    elif data_type in PYTHON_TYPE_TO_AVRO_MAPPING:
        if PYTHON_TYPE_TO_AVRO_MAPPING[data_type] in ["record", "array"]:
            # pylint: disable-next=broad-exception-raised
            raise Exception("Unable to generate Avro schema for dict or array fields without annotation types.")
        avro_type = PYTHON_TYPE_TO_AVRO_MAPPING[data_type]
        field["type"] = avro_type
    elif data_type_origin == list:
        # returns types of list contents
        # if data_type == List[int], arg_data_type = (int,)
        arg_data_type = get_args(data_type)
        if not arg_data_type:
            raise TypeError(
                "List without annotation type is not supported. The argument should be a type, for eg., List[int]"
            )
        avro_type = SIMPLE_PYTHON_TYPE_TO_AVRO_MAPPING.get(arg_data_type[0])
        if avro_type is None:
            raise TypeError(
                "Only following types are supported for list arguments:"
                f" {set(SIMPLE_PYTHON_TYPE_TO_AVRO_MAPPING.keys())}"
            )
        field["type"] = {"type": PYTHON_TYPE_TO_AVRO_MAPPING[data_type_origin], "items": avro_type}
    # Case 3: data_type is an attrs class
    elif hasattr(data_type, "__attrs_attrs__"):
        # Inner Attrs Class

        # fastavro does not allow you to redefine the same record type more than once,
        # so only define an attr record once
        if data_type.__name__ in previously_seen_types:
            field["type"] = data_type.__name__
        else:
            previously_seen_types.add(data_type.__name__)
            record_type = {"name": data_type.__name__, "type": 'record', "fields": []}

            for attribute in data_type.__attrs_attrs__:
                record_type["fields"].append(
                    _create_avro_field_definition(attribute.name, attribute.type,
                                                  previously_seen_types,
                                                  custom_type_to_avro_type=all_field_type_overrides,
                                                  default_is_none=attribute.default is None)
                )
            field["type"] = record_type
    else:
        raise TypeError(
            f"Data type {data_type} is not supported. The data type needs to either"
            " be one of the types in PYTHON_TYPE_TO_AVRO_MAPPING, an attrs decorated class, or one of the types"
            " defined in custom_type_to_avro_type."
        )
    if default_is_none:
        field["default"] = None
        single_type = field["type"]
        field["type"] = ["null", single_type]
    return field



"""A mapping of python types to the Avro type that we want to use make valid avro schema."""
SIMPLE_PYTHON_TYPE_TO_AVRO_MAPPING = {
    bool: "boolean",
    int: "long",
    float: "double",
    bytes: "bytes",
    str: "string",
}
PYTHON_TYPE_TO_AVRO_MAPPING = {
    **SIMPLE_PYTHON_TYPE_TO_AVRO_MAPPING,
    None: "null",
    dict: "record",
    list: "array",
}



"""
Classes and utility functions for the event bus.

This module includes the entry points for the producer and consumer.

API:

- ``get_producer`` returns an ``EventBusProducer`` singleton that should be used for sending all events
  to the Event Bus. The backing implementation is chosen via the Django setting ``EVENT_BUS_PRODUCER``.
  See for example the Kafka implementation's ``KafkaEventProducer``, with the ``create_producer`` function
  serving as the loader: https://github.com/openedx/event-bus-kafka/blob/main/edx_event_bus_kafka/internal/producer.py
- ``make_single_consumer`` creates an ``EventBusConsumer`` instance for a particular combination of
  topic, consumer group, and signal. The backing implementation is chosen via the Django setting
  ``EVENT_BUS_CONSUMER``.
"""

import copy
import warnings
from abc import ABC, abstractmethod
from functools import lru_cache

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

from openedx_events.data import EventsMetadata
from openedx_events.tooling import OpenEdxPublicSignal


def _try_load(*, setting_name: str, args: tuple, kwargs: dict, expected_class: type, default):
    """
    Load an instance of ``expected_class`` as indicated by ``setting_name``.

    The setting points to a callable (function or class) that will fetch or create an
    instance of the expected class. If the configuration is missing or invalid,
    or the callable throws an exception or returns the wrong type, the default is
    returned instead.

    Arguments:
        setting_name: Name of a Django setting containing a dotted module path, indicating a callable
        args: Tuple of positional arguments to pass to the callable
        kwargs: Dictionary of keyword arguments to pass to the callable
        expected_class: The callable must produce an instance of this class object (or a subclass)
        default: Object to return if any part of the lookup or loading fails
    """
    constructor_path = getattr(settings, setting_name, None)
    if constructor_path is None:
        warnings.warn(f"Event Bus setting {setting_name} is missing; component will be inactive")
        return default

    try:
        constructor = import_string(constructor_path)
        instance = constructor(*args, **kwargs)
        if isinstance(instance, expected_class):
            return instance
        else:
            warnings.warn(
                f"{constructor_path} from {setting_name} returned unexpected type {type(instance)}; "
                "component will be inactive"
            )
            return default
    except BaseException as e:
        warnings.warn(
            f"Failed to load {expected_class} from setting {setting_name}: {e!r}; "
            "component will be inactive"
        )
        return default


class EventBusProducer(ABC):
    """
    Parent class for event bus producer implementations.
    """

    @abstractmethod
    def send(
            self, *, signal: OpenEdxPublicSignal, topic: str, event_key_field: str, event_data: dict,
            event_metadata: EventsMetadata
    ) -> None:
        """
        Send a signal event to the event bus under the specified topic.

        Arguments:
            signal: The original OpenEdxPublicSignal the event was sent to
            topic: The event bus topic for the event (without any environmental prefix)
            event_key_field: Path to the event data field to use as the event key (period-delimited
              string naming the dictionary keys to descend)
            event_data: The event data (kwargs) sent to the signal
            event_metadata: The CloudEvent metadata
        """


class NoEventBusProducer(EventBusProducer):
    """
    Stub implementation to "load" when no implementation is properly configured.
    """

    def send(
            self, *, signal: OpenEdxPublicSignal, topic: str, event_key_field: str, event_data: dict,
            event_metadata: EventsMetadata,
    ) -> None:
        """Do nothing."""


# .. setting_name: EVENT_BUS_PRODUCER
# .. setting_default: None
# .. setting_description: String naming a callable (function or class) that can be called to create
#   or retrieve an instance of EventBusProducer when ``openedx_events.event_bus.get_producer`` is
#   called. The format of the string is a dotted path to an attribute in a module, e.g.
#   ``edx_event_bus_kafka.create_producer``. This producer will be managed as a singleton
#   by openedx_events. If setting is not supplied or the callable raises an exception or does not return
#   an instance of EventBusProducer, calls to the producer will be ignored with a warning at startup.

@lru_cache  # will just be one cache entry, in practice
def get_producer() -> EventBusProducer:
    """
    Create or retrieve the producer implementation, as configured.

    If misconfigured, returns a fake implementation that can be called but does nothing.
    """
    return _try_load(
        setting_name='EVENT_BUS_PRODUCER', args=(), kwargs={},
        expected_class=EventBusProducer, default=NoEventBusProducer(),
    )


class EventBusConsumer(ABC):
    """
    Parent class for event bus consumer implementations.
    """

    @abstractmethod
    def consume_indefinitely(self) -> None:
        """
        Consume events from a topic in an infinite loop.

        Events will be converted into calls to Django signals.
        """


class NoEventBusConsumer(EventBusConsumer):
    """
    Stub implementation to "load" when no implementation is properly configured.
    """

    def consume_indefinitely(self) -> None:
        """Do nothing."""


# .. setting_name: EVENT_BUS_CONSUMER
# .. setting_default: None
# .. setting_description: String naming a callable (function or class) that can be called to create
#   or retrieve an instance of EventBusConsumer when ``openedx_events.event_bus.make_single_consumer`` is
#   called. The format of the string is a dotted path to an attribute in a module, e.g.
#   ``some.module.path.EventBusImplementation``. See docstring of ``make_single_consumer`` for
#   parameters. If setting is not supplied or the callable raises an exception or does not return
#   an instance of EventBusConsumer, calls to the consumer will be ignored with a warning at startup.


def make_single_consumer(*, topic: str, group_id: str,
                         **kwargs) -> EventBusConsumer:
    """
    Construct a consumer for a given topic, group, and signal.

    If misconfigured, returns a fake implementation that does nothing.

    Arguments:
        topic: The event bus topic to consume from (without any environmental prefix)
        group_id: The consumer group to participate in
    """
    options = {
        'topic': topic,
        'group_id': group_id,
        **kwargs,
    }
    return _try_load(
        setting_name='EVENT_BUS_CONSUMER', args=(), kwargs=options,
        expected_class=EventBusConsumer, default=NoEventBusConsumer(),
    )


@receiver(setting_changed)
def _reset_state(sender, **kwargs):  # pylint: disable=unused-argument
    """Reset caches when settings change during unit tests."""
    get_producer.cache_clear()


def merge_producer_configs(producer_config_original, producer_config_overrides):
    """
    Merge two EVENT_BUS_PRODUCER_CONFIG maps.

    Arguments:
        producer_config_original: An EVENT_BUS_PRODUCER_CONFIG-structured map
        producer_config_overrides: An EVENT_BUS_PRODUCER_CONFIG-structured map

    Returns:
        A new EVENT_BUS_PRODUCER_CONFIG map created by combining the two maps. All event_type/topic pairs in
        producer_config_overrides are added to the producer_config_original. If there is a conflict on whether a
        particular event_type/topic pair is enabled, producer_config_overrides wins out.
    """
    combined = copy.deepcopy(producer_config_original)
    for event_type, event_type_config_overrides in producer_config_overrides.items():
        event_type_config_combined = combined.get(event_type, {})
        for topic, topic_config_overrides in event_type_config_overrides.items():
            topic_config_combined = event_type_config_combined.get(topic, {})
            enabled_override = topic_config_overrides.get('enabled', None)
            event_key_field_override = topic_config_overrides.get('event_key_field', None)
            if enabled_override is not None:
                topic_config_combined['enabled'] = enabled_override
            if event_key_field_override is not None:
                topic_config_combined['event_key_field'] = event_key_field_override
            event_type_config_combined[topic] = topic_config_combined
        combined[event_type] = event_type_config_combined
    return combined
